"""Tests for ValueRegistry implementations."""

import pytest

from workflow_engine import Value, ValueRegistry
from workflow_engine.core.values.value import (
    ImmutableValueRegistry,
    ValueRegistryBuilder,
)


# Test fixtures - simple value classes for testing
class SampleValueA(Value[int]):
    """Sample value type A."""

    pass


class SampleValueB(Value[str]):
    """Sample value type B."""

    pass


class TestImmutableValueRegistry:
    """Tests for ImmutableValueRegistry."""

    def test_construction(self):
        """Test that ImmutableValueRegistry can be constructed."""
        registry = ImmutableValueRegistry(
            value_classes={"SampleValueA": SampleValueA, "SampleValueB": SampleValueB}
        )
        assert registry is not None

    def test_get_value_class(self):
        """Test retrieving value classes by name."""
        registry = ImmutableValueRegistry(
            value_classes={"SampleValueA": SampleValueA, "SampleValueB": SampleValueB}
        )

        assert registry.get_value_class("SampleValueA") is SampleValueA
        assert registry.get_value_class("SampleValueB") is SampleValueB

    def test_get_value_class_not_found(self):
        """Test that ValueError is raised for missing value types."""
        registry = ImmutableValueRegistry(value_classes={"SampleValueA": SampleValueA})

        with pytest.raises(
            ValueError, match='Value type "NonExistent" is not registered'
        ):
            registry.get_value_class("NonExistent")

    def test_contains_checks_existence(self):
        """Test has_name for checking if a value type is registered."""
        registry = ImmutableValueRegistry(
            value_classes={"SampleValueA": SampleValueA, "SampleValueB": SampleValueB}
        )

        assert registry.has_name("SampleValueA") is True
        assert registry.has_name("SampleValueB") is True
        assert registry.has_name("NonExistent") is False

    def test_getitem_syntax(self):
        """Test get_value_class syntax for retrieving value types."""
        registry = ImmutableValueRegistry(value_classes={"SampleValueA": SampleValueA})

        assert registry.get_value_class("SampleValueA") is SampleValueA

    def test_immutability(self):
        """Test that the registry data cannot be modified after construction."""
        value_classes = {"SampleValueA": SampleValueA}
        registry = ImmutableValueRegistry(value_classes=value_classes)

        # Modifying the original dict should not affect the registry
        value_classes = {"SampleValueA": SampleValueA, "SampleValueB": SampleValueB}

        assert registry.has_name("SampleValueB") is False

    def test_all_value_classes(self):
        """Test that all_value_classes returns all registered value classes."""
        registry = ImmutableValueRegistry(
            value_classes={"SampleValueA": SampleValueA, "SampleValueB": SampleValueB}
        )

        result = dict(registry.all_value_classes())
        assert result == {"SampleValueA": SampleValueA, "SampleValueB": SampleValueB}


class TestEagerValueRegistryBuilder:
    """Tests for EagerValueRegistryBuilder."""

    def test_register_value_class(self):
        """Test registering value classes."""
        builder = ValueRegistry.builder(lazy=False)
        builder.register_value_class(SampleValueA)
        builder.register_value_class(SampleValueB)

        registry = builder.build()
        assert registry.get_value_class("SampleValueA") is SampleValueA
        assert registry.get_value_class("SampleValueB") is SampleValueB

    def test_fluent_interface(self):
        """Test that methods return self for chaining."""
        builder = ValueRegistry.builder(lazy=False)
        result = builder.register_value_class(SampleValueA).register_value_class(
            SampleValueB
        )

        assert result is builder

    def test_duplicate_same_class_succeeds(self):
        """Test that registering the same class twice succeeds (idempotent)."""
        builder = ValueRegistry.builder(lazy=False)
        builder.register_value_class(SampleValueA)
        builder.register_value_class(SampleValueA)  # Same class

        registry = builder.build()
        assert registry.get_value_class("SampleValueA") is SampleValueA

    def test_build_returns_immutable_registry(self):
        """Test that build() returns an ImmutableValueRegistry."""
        builder = ValueRegistry.builder(lazy=False)
        builder.register_value_class(SampleValueA)

        registry = builder.build()
        assert isinstance(registry, ImmutableValueRegistry)
        assert not isinstance(registry, ValueRegistryBuilder)

    def test_register_logs_debug_message(self, caplog):
        """Test that registration logs a debug message."""
        import logging

        caplog.set_level(logging.DEBUG)

        builder = ValueRegistry.builder(lazy=False)
        builder.register_value_class(SampleValueA)

        assert any(
            "Registering value type SampleValueA" in record.message
            for record in caplog.records
        )


class TestLazyValueRegistry:
    """Tests for LazyValueRegistry."""

    def test_starts_unfrozen(self):
        """Test that LazyValueRegistry starts in unfrozen state."""
        registry = ValueRegistry.builder(lazy=True)
        # Should be able to register without error
        registry.register_value_class(SampleValueA)
        assert True  # No exception means it's unfrozen

    def test_register_value_class_when_unfrozen(self):
        """Test registering value classes when unfrozen."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)
        registry.register_value_class(SampleValueB)

        # Trigger build by accessing
        assert registry.get_value_class("SampleValueA") is SampleValueA
        assert registry.get_value_class("SampleValueB") is SampleValueB

    def test_fluent_interface(self):
        """Test that methods return self for chaining."""
        registry = ValueRegistry.builder(lazy=True)
        result = registry.register_value_class(SampleValueA).register_value_class(
            SampleValueB
        )

        assert result is registry

    def test_build_freezes_registry(self):
        """Test that build() freezes the registry."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        # Build should freeze
        registry.build()

        # Should not be able to register after freezing
        with pytest.raises(ValueError, match="Value registry is frozen"):
            registry.register_value_class(SampleValueB)

    def test_access_triggers_build(self):
        """Test that accessing the registry triggers build."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        # Access should trigger build
        registry.get_value_class("SampleValueA")

        # Should be frozen now
        with pytest.raises(ValueError, match="Value registry is frozen"):
            registry.register_value_class(SampleValueB)

    def test_contains_triggers_build(self):
        """Test that has_name triggers build."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        # Access should trigger build
        _ = registry.has_name("SampleValueA")

        # Should be frozen now
        with pytest.raises(ValueError, match="Value registry is frozen"):
            registry.register_value_class(SampleValueB)

    def test_build_is_idempotent(self):
        """Test that calling build() multiple times is safe."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        result1 = registry.build()
        result2 = registry.build()

        assert result1 is result2
        assert result1 is registry

    def test_duplicate_same_class_logs_warning(self, caplog):
        """Test that registering the same class twice logs a warning but succeeds."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)
        registry.register_value_class(SampleValueA)  # Same class

        # Build should log warning but not raise
        registry.build()

        # Check that warning was logged
        assert any(
            "Value type SampleValueA is already registered" in record.message
            for record in caplog.records
        )

    def test_get_value_class_not_found(self):
        """Test that ValueError is raised for missing value types."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        with pytest.raises(
            ValueError, match='Value type "NonExistent" is not registered'
        ):
            registry.get_value_class("NonExistent")

    def test_lazy_registry_is_also_builder(self):
        """Test that LazyValueRegistry is both a registry and a builder."""
        registry = ValueRegistry.builder(lazy=True)

        assert isinstance(registry, ValueRegistry)
        assert isinstance(registry, ValueRegistryBuilder)

    def test_build_returns_self(self):
        """Test that build() returns self (the LazyValueRegistry itself)."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        result = registry.build()
        assert result is registry
        assert isinstance(result, ValueRegistry)

    def test_registrations_deleted_after_build(self):
        """Test that _registrations list is deleted after build to save memory."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        assert hasattr(registry, "_registrations")
        registry.build()

        # _registrations should be deleted
        assert not hasattr(registry, "_registrations")

    def test_getitem_syntax(self):
        """Test get_value_class syntax for retrieving value types."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        assert registry.get_value_class("SampleValueA") is SampleValueA

    def test_all_value_classes(self):
        """Test that all_value_classes returns all registered value classes."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)
        registry.register_value_class(SampleValueB)

        result = dict(registry.all_value_classes())
        assert result == {"SampleValueA": SampleValueA, "SampleValueB": SampleValueB}

    def test_all_value_classes_triggers_build(self):
        """Test that all_value_classes triggers build."""
        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        _ = list(registry.all_value_classes())

        # Should be frozen now
        with pytest.raises(ValueError, match="Value registry is frozen"):
            registry.register_value_class(SampleValueB)


class TestValueRegistryIntegration:
    """Integration tests with actual Value classes."""

    def test_global_registry_contains_builtin_types(self):
        """Test that the global default_value_registry contains built-in value types."""
        # These types should have been auto-registered during import
        assert ValueRegistry.DEFAULT.has_name("IntegerValue")
        assert ValueRegistry.DEFAULT.has_name("StringValue")
        assert ValueRegistry.DEFAULT.has_name("FloatValue")
        assert ValueRegistry.DEFAULT.has_name("BooleanValue")

    def test_retrieve_builtin_types_from_global_registry(self):
        """Test retrieving built-in value types from global registry."""
        from workflow_engine.core.values import (
            BooleanValue,
            FloatValue,
            IntegerValue,
            StringValue,
        )

        assert ValueRegistry.DEFAULT.get_value_class("IntegerValue") is IntegerValue
        assert ValueRegistry.DEFAULT.get_value_class("StringValue") is StringValue
        assert ValueRegistry.DEFAULT.get_value_class("FloatValue") is FloatValue
        assert ValueRegistry.DEFAULT.get_value_class("BooleanValue") is BooleanValue

    def test_with_concrete_value_classes(self):
        """Test registry with real value classes."""
        registry = ValueRegistry.builder(lazy=True)

        # Create isolated registry for testing
        registry.register_value_class(SampleValueA)
        registry.register_value_class(SampleValueB)

        # Retrieve and verify
        assert registry.get_value_class("SampleValueA") is SampleValueA
        assert registry.get_value_class("SampleValueB") is SampleValueB

    def test_mixed_registration(self):
        """Test registry with multiple value classes."""
        registry = ValueRegistry.builder(lazy=True)

        registry.register_value_class(SampleValueA)
        registry.register_value_class(SampleValueB)

        registry.build()

        assert registry.has_name("SampleValueA")
        assert registry.get_value_class("SampleValueA") is SampleValueA
        assert registry.has_name("SampleValueB")
        assert registry.get_value_class("SampleValueB") is SampleValueB


class TestValueRegistryEdgeCases:
    """Edge case tests for ValueRegistry."""

    def test_multiple_registries_are_independent(self):
        """Test that multiple LazyValueRegistry instances are independent."""
        registry1 = ValueRegistry.builder(lazy=True)
        registry2 = ValueRegistry.builder(lazy=True)

        registry1.register_value_class(SampleValueA)
        registry2.register_value_class(SampleValueB)

        assert registry1.has_name("SampleValueA")
        assert registry1.get_value_class("SampleValueA") is SampleValueA
        assert registry2.has_name("SampleValueB")
        assert registry2.get_value_class("SampleValueB") is SampleValueB
        assert registry2.has_name("SampleValueA") is False
        with pytest.raises(
            ValueError, match='Value type "SampleValueA" is not registered'
        ):
            registry2.get_value_class("SampleValueA")


class TestValueRegistryExtend:
    """Tests for ValueRegistry.extend() method."""

    def test_extend_eager(self):
        """Test extending a registry into an eager builder."""
        registry = ImmutableValueRegistry(value_classes={"SampleValueA": SampleValueA})

        builder = registry.extend(lazy=False)
        new_registry = builder.build()

        assert new_registry.get_value_class("SampleValueA") is SampleValueA

    def test_extend_lazy(self):
        """Test extending a registry into a lazy builder."""
        registry = ImmutableValueRegistry(value_classes={"SampleValueA": SampleValueA})

        lazy_registry = registry.extend(lazy=True)
        assert lazy_registry.get_value_class("SampleValueA") is SampleValueA

    def test_extend_preserves_all_classes(self):
        """Test that extend copies all registered classes."""
        registry = ImmutableValueRegistry(
            value_classes={"SampleValueA": SampleValueA, "SampleValueB": SampleValueB}
        )

        builder = registry.extend(lazy=False)
        new_registry = builder.build()

        assert new_registry.get_value_class("SampleValueA") is SampleValueA
        assert new_registry.get_value_class("SampleValueB") is SampleValueB


class TestValueRegistryLoadValue:
    """Tests for ValueRegistry.load_value() method."""

    def test_load_value_with_matching_title(self):
        """Test loading a value type from schema with matching title (backwards compat)."""
        from workflow_engine.core.values.schema import IntegerValueSchema

        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        schema = IntegerValueSchema(type="integer", title="SampleValueA")
        loaded_value = registry.load_value(schema)

        assert loaded_value is SampleValueA

    def test_load_value_with_matching_value_type(self):
        """Test loading a value type from schema with matching x-value-type (preferred)."""
        from workflow_engine.core.values.schema import IntegerValueSchema

        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        schema = IntegerValueSchema.model_validate(
            {"type": "integer", "x-value-type": "SampleValueA"}
        )
        loaded_value = registry.load_value(schema)

        assert loaded_value is SampleValueA

    def test_load_value_prefers_value_type_over_title(self):
        """Test that x-value-type takes priority over title when both are present."""
        from workflow_engine.core.values.schema import IntegerValueSchema

        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)
        registry.register_value_class(SampleValueB)

        schema = IntegerValueSchema.model_validate(
            {
                "type": "integer",
                "x-value-type": "SampleValueA",
                "title": "SampleValueB",
            }
        )
        loaded_value = registry.load_value(schema)

        assert loaded_value is SampleValueA

    def test_load_value_without_title_returns_none(self):
        """Test that load_value returns None when schema has no title."""
        from workflow_engine.core.values.schema import IntegerValueSchema

        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        schema = IntegerValueSchema(type="integer")  # No title
        loaded_value = registry.load_value(schema)

        assert loaded_value is None

    def test_load_value_with_unregistered_title_returns_none(self):
        """Test that load_value returns None for unregistered title."""
        from workflow_engine.core.values.schema import IntegerValueSchema

        registry = ValueRegistry.builder(lazy=True)
        registry.register_value_class(SampleValueA)

        schema = IntegerValueSchema(type="integer", title="UnregisteredType")
        loaded_value = registry.load_value(schema)

        assert loaded_value is None

    def test_load_value_integration_with_to_value_cls(self):
        """Test that ValueSchema.to_value_cls() uses the registry's load_value."""
        from workflow_engine.core.values import IntegerValue
        from workflow_engine.core.values.schema import IntegerValueSchema

        # Create a schema with a title that matches a built-in type
        schema = IntegerValueSchema(type="integer", title="IntegerValue")

        # to_value_cls should use the default registry to load the value
        value_cls = schema.to_value_cls()

        assert value_cls is IntegerValue



"""Tests for NodeRegistry implementations."""

from functools import cached_property
from typing import Literal

import pytest

from workflow_engine import (
    Empty,
    Node,
    NodeRegistry,
    NodeTypeInfo,
)
from workflow_engine.core.node import ImmutableNodeRegistry, NodeRegistryBuilder


# Test fixtures - simple node classes for testing
class SampleNodeA(Node[Empty, Empty, Empty]):
    TYPE_INFO = NodeTypeInfo.from_parameter_type(
        name="TestA",
        display_name="Test A",
        version="1.0.0",
        parameter_type=Empty,
    )
    type: Literal["TestA"] = "TestA"  # pyright: ignore[reportIncompatibleVariableOverride]

    @cached_property
    def input_type(self):
        return Empty

    @cached_property
    def output_type(self):
        return Empty

    async def run(self, context, input):
        return Empty()


class SampleNodeB(Node[Empty, Empty, Empty]):
    TYPE_INFO = NodeTypeInfo.from_parameter_type(
        name="TestB",
        display_name="Test B",
        version="1.0.0",
        parameter_type=Empty,
    )
    type: Literal["TestB"] = "TestB"  # pyright: ignore[reportIncompatibleVariableOverride]

    @cached_property
    def input_type(self):
        return Empty

    @cached_property
    def output_type(self):
        return Empty

    async def run(self, context, input):
        return Empty()


class SampleBaseNode(Node[Empty, Empty, Empty]):
    """A base node class without a type discriminator."""

    pass


class TestImmutableNodeRegistry:
    """Tests for ImmutableNodeRegistry."""

    def test_construction(self):
        """Test that ImmutableNodeRegistry can be constructed."""
        registry = ImmutableNodeRegistry(
            node_classes={"TestA": SampleNodeA, "TestB": SampleNodeB},
            base_node_classes=[Node, SampleBaseNode],
        )
        assert registry is not None

    def test_get_node_class(self):
        """Test retrieving node classes by name."""
        registry = ImmutableNodeRegistry(
            node_classes={"TestA": SampleNodeA, "TestB": SampleNodeB},
            base_node_classes=[],
        )

        assert registry.get_node_class("TestA") is SampleNodeA
        assert registry.get_node_class("TestB") is SampleNodeB

    def test_get_node_class_not_found(self):
        """Test that KeyError is raised for missing node types."""
        registry = ImmutableNodeRegistry(
            node_classes={"TestA": SampleNodeA},
            base_node_classes=[],
        )

        with pytest.raises(KeyError):
            registry.get_node_class("NonExistent")

    def test_is_base_node_class(self):
        """Test checking if a class is registered as a base class."""
        registry = ImmutableNodeRegistry(
            node_classes={},
            base_node_classes=[Node, SampleBaseNode],
        )

        assert registry.is_base_node_class(Node)
        assert registry.is_base_node_class(SampleBaseNode)
        assert not registry.is_base_node_class(SampleNodeA)

    def test_immutability(self):
        """Test that the registry data cannot be modified after construction."""
        node_classes: dict[str, type[Node]] = {"TestA": SampleNodeA}
        base_node_classes = [Node]

        registry = ImmutableNodeRegistry(
            node_classes=node_classes,
            base_node_classes=base_node_classes,
        )

        # Modifying the original dicts should not affect the registry
        node_classes["TestB"] = SampleNodeB
        base_node_classes.append(SampleBaseNode)

        with pytest.raises(KeyError):
            registry.get_node_class("TestB")
        assert not registry.is_base_node_class(SampleBaseNode)


class TestEagerNodeRegistryBuilder:
    """Tests for EagerNodeRegistryBuilder."""

    def test_register_node_class(self):
        """Test registering node classes."""
        registry = (
            NodeRegistry.builder(lazy=False)
            .register_node_class(SampleNodeA)
            .register_node_class(SampleNodeB)
            .build()
        )
        assert registry.get_node_class("TestA") is SampleNodeA
        assert registry.get_node_class("TestB") is SampleNodeB

    def test_register_base_node_class(self):
        """Test registering base node classes."""
        registry = (
            NodeRegistry.builder(lazy=False)
            .register_node_class(Node)
            .register_node_class(SampleBaseNode)
            .build()
        )
        assert registry.is_base_node_class(Node)
        assert registry.is_base_node_class(SampleBaseNode)

    def test_fluent_interface(self):
        """Test that methods return self for chaining."""
        builder1 = NodeRegistry.builder(lazy=False)
        builder2 = (
            builder1.register_node_class(SampleNodeA)
            .register_node_class(SampleNodeB)
            .register_node_class(Node)
        )
        assert builder1 is builder2

    def test_duplicate_node_class_raises_error(self):
        """Test that registering the same node type twice raises ValueError."""
        builder = NodeRegistry.builder(lazy=False)
        builder.register_node_class(SampleNodeA)

        with pytest.raises(ValueError, match='Node type "TestA" is already registered'):
            builder.register_node_class(SampleNodeA)

    def test_build_returns_immutable_registry(self):
        """Test that build() returns an ImmutableNodeRegistry."""
        builder = NodeRegistry.builder(lazy=False)
        builder.register_node_class(SampleNodeA)

        registry = builder.build()
        assert isinstance(registry, ImmutableNodeRegistry)
        assert not isinstance(registry, NodeRegistryBuilder)


class TestLazyNodeRegistry:
    """Tests for LazyNodeRegistry."""

    def test_starts_unfrozen(self):
        """Test that LazyNodeRegistry starts in unfrozen state."""
        registry = NodeRegistry.builder(lazy=True)
        # Should be able to register without error
        registry.register_node_class(SampleNodeA)
        assert True  # No exception means it's unfrozen

    def test_register_node_class_when_unfrozen(self):
        """Test registering node classes when unfrozen."""
        (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleNodeA)
            .register_node_class(SampleNodeB)
            .build()
        )

    def test_fluent_interface(self):
        """Test that methods return self for chaining."""
        registry = NodeRegistry.builder(lazy=True)
        result = (
            registry.register_node_class(SampleNodeB)
            .register_node_class(Node)
            .register_node_class(SampleNodeA)
        )

        assert result is registry

    def test_build_freezes_registry(self):
        """Test that build() freezes the registry."""
        registry = NodeRegistry.builder(lazy=True)
        registry.register_node_class(SampleNodeA)

        # Build should freeze
        registry.build()

        # Should not be able to register after freezing
        with pytest.raises(ValueError, match="Node registry is frozen"):
            registry.register_node_class(SampleNodeB)

    def test_access_triggers_build(self):
        """Test that accessing the registry triggers build."""
        registry = NodeRegistry.builder(lazy=True)
        registry.register_node_class(SampleNodeA)

        # Access should trigger build
        registry.get_node_class("TestA")

        # Should be frozen now
        with pytest.raises(ValueError, match="Node registry is frozen"):
            registry.register_node_class(SampleNodeB)

    def test_is_base_node_class_triggers_build(self):
        """Test that is_base_node_class triggers build."""
        registry = NodeRegistry.builder(lazy=True)
        registry.register_node_class(Node)

        # Access should trigger build
        registry.is_base_node_class(Node)

        # Should be frozen now
        with pytest.raises(ValueError, match="Node registry is frozen"):
            registry.register_node_class(SampleNodeA)

    def test_build_is_idempotent(self):
        """Test that calling build() multiple times is safe."""
        registry = NodeRegistry.builder(lazy=True)
        registry.register_node_class(SampleNodeA)

        result1 = registry.build()
        result2 = registry.build()

        assert result1 is result2
        assert result1 is registry

    def test_duplicate_same_class_logs_warning(self, caplog):
        """Test that registering the same class twice logs a warning but succeeds."""
        registry = NodeRegistry.builder(lazy=True)
        registry.register_node_class(SampleNodeA)
        registry.register_node_class(SampleNodeA)  # Same class

        # Build should log warning but not raise
        registry.build()

        # Check that warning was logged
        assert any(
            "Node type TestA is already registered" in record.message
            for record in caplog.records
        )

    def test_duplicate_different_class_raises_error(self):
        """Test that registering a different class with the same type raises ValueError."""
        # Create a duplicate class inside the test, patching DEFAULT to avoid
        # polluting it (since __init_subclass__ registers into DEFAULT).
        original_default = NodeRegistry.DEFAULT
        NodeRegistry.DEFAULT = NodeRegistry.builder(lazy=True)
        try:

            class DuplicateTestA(Node[Empty, Empty, Empty]):
                TYPE_INFO = NodeTypeInfo.from_parameter_type(
                    name="TestA",
                    display_name="Test A Dup",
                    version="1.0.0",
                    parameter_type=Empty,
                )
                type: Literal["TestA"] = "TestA"  # pyright: ignore[reportIncompatibleVariableOverride]

                @cached_property
                def input_type(self):
                    return Empty

                @cached_property
                def output_type(self):
                    return Empty

                async def run(self, context, input):
                    return Empty()

        finally:
            NodeRegistry.DEFAULT = original_default

        registry = NodeRegistry.builder(lazy=True)
        registry.register_node_class(SampleNodeA)
        registry.register_node_class(DuplicateTestA)

        # Build should raise error
        with pytest.raises(
            ValueError,
            match='Node type "TestA" .* is already registered to a different class',
        ):
            registry.build()

    def test_duplicate_base_class_logs_warning(self, caplog):
        """Test that registering the same base class twice logs a warning."""
        registry = NodeRegistry.builder(lazy=True)
        registry.register_node_class(Node)
        registry.register_node_class(Node)

        registry.build()

        assert any(
            "Node base class Node is already registered" in record.message
            for record in caplog.records
        )

    def test_get_node_class_not_found(self):
        """Test that ValueError is raised for missing node types."""
        registry = NodeRegistry.builder(lazy=True)
        registry.register_node_class(SampleNodeA)

        with pytest.raises(
            ValueError, match='Node type "NonExistent" is not registered'
        ):
            registry.get_node_class("NonExistent")

    def test_lazy_registry_is_also_registry(self):
        """Test that LazyNodeRegistry is both a builder and a registry."""
        registry = NodeRegistry.builder(lazy=True)

        assert isinstance(registry, NodeRegistry)
        assert isinstance(registry, NodeRegistryBuilder)

    def test_build_returns_self(self):
        """Test that build() returns self (the LazyNodeRegistry itself)."""
        registry = NodeRegistry.builder(lazy=True)
        registry.register_node_class(SampleNodeA)

        result = registry.build()
        assert result is registry
        assert isinstance(result, NodeRegistry)

    def test_registrations_deleted_after_build(self):
        """Test that _registrations list is deleted after build to save memory."""
        registry = NodeRegistry.builder(lazy=True)
        registry.register_node_class(SampleNodeA)

        assert hasattr(registry, "_registrations")
        registry.build()

        # _registrations should be deleted
        assert not hasattr(registry, "_registrations")


class TestNodeRegistryIntegration:
    """Integration tests with actual Node classes."""

    def test_with_concrete_node_classes(self):
        """Test registry with real node classes that have type discriminators."""
        registry = NodeRegistry.builder(lazy=True)

        # These classes should have been registered via __init_subclass__
        # but we're creating a fresh registry for isolation
        registry.register_node_class(SampleNodeA)
        registry.register_node_class(SampleNodeB)

        # Retrieve and verify
        assert registry.get_node_class("TestA") is SampleNodeA
        assert registry.get_node_class("TestB") is SampleNodeB

    def test_with_base_node_classes(self):
        """Test registry with base node classes (no type discriminator)."""
        registry = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleBaseNode)
            .build()
        )

        assert registry.is_base_node_class(Node)
        assert registry.is_base_node_class(SampleBaseNode)
        assert not registry.is_base_node_class(SampleNodeA)  # Has discriminator

    def test_mixed_registration(self):
        """Test registry with both concrete and base classes."""
        registry = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(SampleNodeA)
            .register_node_class(SampleBaseNode)
            .build()
        )

        assert registry.get_node_class("TestA") is SampleNodeA
        assert registry.is_base_node_class(SampleBaseNode)
        assert not registry.is_base_node_class(SampleNodeA)


class TestNodeRegistryLoadNode:
    """Tests for NodeRegistry.load_node() method."""

    def test_load_node_with_base_class(self):
        """Test loading a base node instance into concrete type."""
        registry = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleNodeA)
            .build()
        )

        # Create a base Node instance using model_construct to bypass validators
        base_node = Node.model_construct(type="TestA", id="test-1", version="1.0.0")

        # Load it into the concrete type
        loaded_node = registry.load_node(base_node)

        assert isinstance(loaded_node, SampleNodeA)
        assert loaded_node.type == "TestA"
        assert loaded_node.id == "test-1"
        assert loaded_node.version == "1.0.0"

    def test_load_node_with_concrete_class_returns_unchanged(self):
        """Test that loading a concrete node returns it unchanged."""
        registry = (
            NodeRegistry.builder(lazy=True).register_node_class(SampleNodeA).build()
        )

        # Create a concrete node instance
        concrete_node = SampleNodeA(type="TestA", id="test-1", version="1.0.0")

        # Load should return the same instance
        loaded_node = registry.load_node(concrete_node)

        assert loaded_node is concrete_node
        assert isinstance(loaded_node, SampleNodeA)

    def test_load_node_unregistered_type_raises_error(self):
        """Test that loading a node with unregistered type raises ValueError."""
        registry = NodeRegistry.builder(lazy=True).register_node_class(Node).build()

        # Create a base Node with unregistered type using model_construct to bypass validation
        base_node = Node.model_construct(
            type="UnregisteredType", id="test-1", version="1.0.0"
        )

        with pytest.raises(
            ValueError, match='Node type "UnregisteredType" is not registered'
        ):
            registry.load_node(base_node)

    def test_load_node_with_different_registries(self):
        """Test that different registries load to different concrete types."""
        registry_a = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleNodeA)
            .build()
        )

        registry_b = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleNodeB)
            .build()
        )

        base_node_a = Node.model_construct(type="TestA", id="test-1", version="1.0.0")
        base_node_b = Node.model_construct(type="TestB", id="test-2", version="1.0.0")

        loaded_a = registry_a.load_node(base_node_a)
        loaded_b = registry_b.load_node(base_node_b)

        assert isinstance(loaded_a, SampleNodeA)
        assert isinstance(loaded_b, SampleNodeB)

        # Cross-registry loading should fail
        with pytest.raises(ValueError, match='Node type "TestB" is not registered'):
            registry_a.load_node(base_node_b)

    def test_load_node_preserves_params(self):
        """Test that loading preserves node parameters."""
        registry = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(SampleNodeA)
            .register_node_class(Node)
            .build()
        )

        base_node = Node.model_construct(
            type="TestA",
            id="test-1",
            version="1.0.0",
            params=Empty(),
        )

        loaded_node = registry.load_node(base_node)

        assert isinstance(loaded_node, SampleNodeA)
        assert loaded_node.params is not None
        assert isinstance(loaded_node.params, Empty)



"""Tests for WorkflowEngine."""

from functools import cached_property
from typing import Literal

import pytest

from workflow_engine import (
    Edge,
    Empty,
    IntegerValue,
    Node,
    NodeRegistry,
    NodeTypeInfo,
    ValueRegistry,
    Workflow,
    WorkflowEngine,
    WorkflowExecutionResultStatus,
)
from workflow_engine.contexts import InMemoryContext
from workflow_engine.core.io import InputNode, OutputNode
from workflow_engine.execution import ParallelExecutionAlgorithm
from workflow_engine.nodes import AddNode


# Test fixtures - simple node classes for testing
class SampleAddNode(Node[Empty, Empty, Empty]):
    TYPE_INFO = NodeTypeInfo.from_parameter_type(
        name="SampleAdd",
        display_name="Sample Add",
        version="1.0.0",
        parameter_type=Empty,
    )
    type: Literal["SampleAdd"] = "SampleAdd"  # pyright: ignore[reportIncompatibleVariableOverride]

    @cached_property
    def input_type(self):
        return Empty

    @cached_property
    def output_type(self):
        return Empty

    async def run(self, context, input):
        return Empty()


class SampleMultiplyNode(Node[Empty, Empty, Empty]):
    TYPE_INFO = NodeTypeInfo.from_parameter_type(
        name="SampleMultiply",
        display_name="Sample Multiply",
        version="1.0.0",
        parameter_type=Empty,
    )
    type: Literal["SampleMultiply"] = "SampleMultiply"  # pyright: ignore[reportIncompatibleVariableOverride]

    @cached_property
    def input_type(self):
        return Empty

    @cached_property
    def output_type(self):
        return Empty

    async def run(self, context, input):
        return Empty()


class TestWorkflowEngine:
    """Tests for WorkflowEngine class."""

    def test_construction(self):
        """Test that WorkflowEngine can be constructed."""
        node_registry = NodeRegistry.builder(lazy=True)
        value_registry = ValueRegistry.builder(lazy=True)

        engine = WorkflowEngine(
            node_registry=node_registry,
            value_registry=value_registry,
        )

        assert engine.node_registry is node_registry
        assert engine.value_registry is value_registry

    def test_construction_with_defaults(self):
        """Test that WorkflowEngine uses default registries when none provided."""
        engine = WorkflowEngine()

        assert engine.node_registry is NodeRegistry.DEFAULT
        assert engine.value_registry is ValueRegistry.DEFAULT

    def test_default_engine_loads_workflow(self):
        """Test that default engine can load workflows with built-in node types."""
        # Create engine with no arguments - uses global registries
        engine = WorkflowEngine()

        # Use SampleAddNode which is registered in our test's global registry
        input_node = InputNode.empty()
        output_node = OutputNode.empty()

        workflow = Workflow.model_construct(
            input_node=input_node,
            inner_nodes=[
                Node.model_construct(type="SampleAdd", id="node1", version="1.0.0")
            ],
            output_node=output_node,
            edges=[],
        )

        # Should be able to load it
        typed_workflow = engine.load(workflow)
        assert isinstance(typed_workflow.inner_nodes[0], SampleAddNode)

    def test_load_workflow_with_untyped_nodes(self):
        """Test loading a workflow with untyped nodes into typed nodes."""
        # Set up registry
        node_registry = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleAddNode)
            .build()
        )

        value_registry = ValueRegistry.builder(lazy=True)

        engine = WorkflowEngine(
            node_registry=node_registry,
            value_registry=value_registry,
        )

        # Create workflow with untyped node
        input_node = InputNode.empty()
        output_node = OutputNode.empty()
        untyped_node = Node.model_construct(
            type="SampleAdd", id="node1", version="1.0.0"
        )
        workflow = Workflow(
            input_node=input_node,
            inner_nodes=[untyped_node],
            output_node=output_node,
            edges=[],
        )

        # Load workflow
        typed_workflow = engine.load(workflow)

        # Verify the node is now typed
        assert len(typed_workflow.inner_nodes) == 1
        assert isinstance(typed_workflow.inner_nodes[0], SampleAddNode)
        assert typed_workflow.inner_nodes[0].type == "SampleAdd"
        assert typed_workflow.inner_nodes[0].id == "node1"

    def test_load_workflow_with_already_typed_nodes(self):
        """Test that loading a workflow with typed nodes returns equivalent workflow."""
        node_registry = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleAddNode)
            .build()
        )

        value_registry = ValueRegistry.builder(lazy=True)

        engine = WorkflowEngine(
            node_registry=node_registry,
            value_registry=value_registry,
        )

        # Create workflow with already typed node
        input_node = InputNode.empty()
        output_node = OutputNode.empty()
        typed_node = SampleAddNode(
            id="node1",
            version="1.0.0",
        )
        workflow = Workflow(
            input_node=input_node,
            inner_nodes=[typed_node],
            output_node=output_node,
            edges=[],
        )

        # Load workflow
        loaded_workflow = engine.load(workflow)

        # Verify the node is still the same type
        assert len(loaded_workflow.inner_nodes) == 1
        assert isinstance(loaded_workflow.inner_nodes[0], SampleAddNode)
        assert loaded_workflow.inner_nodes[0].id == "node1"

    def test_load_workflow_with_multiple_nodes(self):
        """Test loading a workflow with multiple untyped nodes."""
        node_registry = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleAddNode)
            .register_node_class(SampleMultiplyNode)
            .build()
        )

        value_registry = ValueRegistry.builder(lazy=True)

        engine = WorkflowEngine(
            node_registry=node_registry,
            value_registry=value_registry,
        )

        # Create workflow with multiple untyped nodes
        input_node = InputNode.empty()
        output_node = OutputNode.empty()
        node1 = Node.model_construct(type="SampleAdd", id="node1", version="1.0.0")
        node2 = Node.model_construct(type="SampleMultiply", id="node2", version="1.0.0")

        workflow = Workflow(
            input_node=input_node,
            inner_nodes=[node1, node2],
            output_node=output_node,
            edges=[],
        )

        # Load workflow
        typed_workflow = engine.load(workflow)

        # Verify both nodes are typed
        assert len(typed_workflow.inner_nodes) == 2
        assert isinstance(typed_workflow.inner_nodes[0], SampleAddNode)
        assert isinstance(typed_workflow.inner_nodes[1], SampleMultiplyNode)

    def test_load_workflow_with_unregistered_node_type_raises_error(self):
        """Test that loading a workflow with unregistered node type raises ValueError."""
        node_registry = (
            NodeRegistry.builder(lazy=True).register_node_class(Node).build()
        )

        value_registry = ValueRegistry.builder(lazy=True)

        engine = WorkflowEngine(
            node_registry=node_registry,
            value_registry=value_registry,
        )

        # Create workflow with unregistered node type using model_construct to bypass validation
        input_node = InputNode.empty()
        output_node = OutputNode.empty()
        untyped_node = Node.model_construct(
            type="UnregisteredType",
            id="node1",
            version="1.0.0",
        )
        workflow = Workflow.model_construct(
            input_node=input_node,
            inner_nodes=[untyped_node],
            output_node=output_node,
            edges=[],
        )

        # Loading should raise error
        with pytest.raises(
            ValueError, match='Node type "UnregisteredType" is not registered'
        ):
            engine.load(workflow)

    def test_load_workflow_with_different_registries(self):
        """Test that different engines with different registries have different capabilities."""
        # Engine A with both SampleAddNode and SampleMultiplyNode
        node_registry_a = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleAddNode)
            .register_node_class(SampleMultiplyNode)
            .build()
        )

        engine_a = WorkflowEngine(
            node_registry=node_registry_a,
            value_registry=ValueRegistry.builder(lazy=True),
        )

        # Engine B with only SampleAddNode
        node_registry_b = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleAddNode)
            .build()
        )

        engine_b = WorkflowEngine(
            node_registry=node_registry_b,
            value_registry=ValueRegistry.builder(lazy=True),
        )

        # Workflow with SampleAdd node - both can load
        input_node = InputNode.empty()
        output_node = OutputNode.empty()

        workflow_add = Workflow.model_construct(
            input_node=input_node,
            inner_nodes=[
                Node.model_construct(type="SampleAdd", id="node1", version="1.0.0")
            ],
            output_node=output_node,
            edges=[],
        )

        workflow_a = engine_a.load(workflow_add)
        workflow_b = engine_b.load(workflow_add)

        assert isinstance(workflow_a.inner_nodes[0], SampleAddNode)
        assert isinstance(workflow_b.inner_nodes[0], SampleAddNode)

        # Workflow with SampleMultiply node - only engine A can load
        workflow_multiply = Workflow.model_construct(
            input_node=input_node,
            inner_nodes=[
                Node.model_construct(type="SampleMultiply", id="node1", version="1.0.0")
            ],
            output_node=output_node,
            edges=[],
        )

        workflow_a_mult = engine_a.load(workflow_multiply)
        assert isinstance(workflow_a_mult.inner_nodes[0], SampleMultiplyNode)

        # Engine B cannot load SampleMultiply
        with pytest.raises(
            ValueError, match='Node type "SampleMultiply" is not registered'
        ):
            engine_b.load(workflow_multiply)

    def test_load_preserves_workflow_structure(self):
        """Test that loading preserves edges and other workflow structure."""
        node_registry = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleAddNode)
            .build()
        )

        value_registry = ValueRegistry.builder(lazy=True)

        engine = WorkflowEngine(
            node_registry=node_registry,
            value_registry=value_registry,
        )

        # Create workflow with multiple nodes (no edges for simplicity)
        input_node = InputNode.empty()
        output_node = OutputNode.empty()
        node1 = Node.model_construct(type="SampleAdd", id="node1", version="1.0.0")
        node2 = Node.model_construct(type="SampleAdd", id="node2", version="1.0.0")

        workflow = Workflow.model_construct(
            input_node=input_node,
            inner_nodes=[node1, node2],
            output_node=output_node,
            edges=[],
        )

        # Load workflow
        typed_workflow = engine.load(workflow)

        # Verify structure is preserved
        assert len(typed_workflow.inner_nodes) == 2
        assert typed_workflow.input_node.id == "input"
        assert typed_workflow.output_node.id == "output"
        assert typed_workflow.inner_nodes[0].id == "node1"
        assert typed_workflow.inner_nodes[1].id == "node2"


class TestWorkflowEngineMultiTenancy:
    """Tests for multi-tenancy scenarios with WorkflowEngine."""

    def test_tenant_isolation(self):
        """Test that different tenants can have different node types available."""
        # Tenant A has only SampleAddNode
        tenant_a_registry = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleAddNode)
            .build()
        )

        engine_a = WorkflowEngine(
            node_registry=tenant_a_registry,
        )

        # Tenant B has only SampleMultiplyNode
        tenant_b_registry = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleMultiplyNode)
            .build()
        )

        engine_b = WorkflowEngine(
            node_registry=tenant_b_registry,
        )

        # Workflow with SampleAdd node - use model_construct to bypass validation
        input_node = InputNode.empty()
        output_node = OutputNode.empty()

        workflow_add = Workflow.model_construct(
            input_node=input_node,
            inner_nodes=[
                Node.model_construct(type="SampleAdd", id="node1", version="1.0.0")
            ],
            output_node=output_node,
            edges=[],
        )

        # Tenant A can load it
        loaded_a = engine_a.load(workflow_add)
        assert isinstance(loaded_a.inner_nodes[0], SampleAddNode)

        # Tenant B cannot load it
        with pytest.raises(ValueError, match='Node type "SampleAdd" is not registered'):
            engine_b.load(workflow_add)

    def test_shared_workflow_different_engines(self):
        """Test that the same workflow JSON can be loaded by different engines."""
        # Create two engines with different registries
        registry_1 = NodeRegistry.builder(lazy=True)
        registry_1.register_node_class(SampleAddNode)
        registry_1.register_node_class(Node)

        registry_2 = NodeRegistry.builder(lazy=True)
        registry_2.register_node_class(SampleAddNode)
        registry_2.register_node_class(Node)

        engine_1 = WorkflowEngine(
            node_registry=registry_1,
            value_registry=ValueRegistry.builder(lazy=True),
        )

        engine_2 = WorkflowEngine(
            node_registry=registry_2,
            value_registry=ValueRegistry.builder(lazy=True),
        )

        # Create a workflow
        input_node = InputNode.empty()
        output_node = OutputNode.empty()

        workflow = Workflow(
            input_node=input_node,
            inner_nodes=[
                Node.model_construct(type="SampleAdd", id="node1", version="1.0.0")
            ],
            output_node=output_node,
            edges=[],
        )

        # Both engines can load the same workflow independently
        loaded_1 = engine_1.load(workflow)
        loaded_2 = engine_2.load(workflow)

        # Both should produce typed workflows
        assert isinstance(loaded_1.inner_nodes[0], SampleAddNode)
        assert isinstance(loaded_2.inner_nodes[0], SampleAddNode)


class TestWorkflowEngineExecution:
    """Tests for WorkflowEngine.execute() method."""

    @pytest.mark.asyncio
    async def test_execute_with_default_algorithm(self):
        """Test that execute() works with default TopologicalExecutionAlgorithm."""

        # Create engine with defaults
        engine = WorkflowEngine()

        # Create a simple addition workflow
        input_node = InputNode.from_fields(
            a=IntegerValue,
            b=IntegerValue,
        )
        output_node = OutputNode.from_fields(
            result=IntegerValue,
        )
        add_node = AddNode(id="add1")

        workflow = Workflow(
            input_node=input_node,
            inner_nodes=[add_node],
            output_node=output_node,
            edges=[
                Edge.from_nodes(
                    source=input_node, source_key="a", target=add_node, target_key="a"
                ),
                Edge.from_nodes(
                    source=input_node, source_key="b", target=add_node, target_key="b"
                ),
                Edge.from_nodes(
                    source=add_node,
                    source_key="sum",
                    target=output_node,
                    target_key="result",
                ),
            ],
        )

        # Execute workflow
        context = InMemoryContext()
        input_data = {"a": IntegerValue(5), "b": IntegerValue(3)}

        result = await engine.execute(workflow, input_data, context)

        # Verify results
        assert result.status is WorkflowExecutionResultStatus.SUCCESS
        assert result.output["result"].root == 8

    @pytest.mark.asyncio
    async def test_execute_with_custom_algorithm(self):
        """Test that execute() uses the provided execution algorithm."""

        # Create engine with custom algorithm
        custom_algorithm = ParallelExecutionAlgorithm()
        engine = WorkflowEngine(execution_algorithm=custom_algorithm)

        # Verify the algorithm is stored
        assert engine.execution_algorithm is custom_algorithm

        # Create a simple workflow
        input_node = InputNode.from_fields(
            a=IntegerValue,
            b=IntegerValue,
        )
        output_node = OutputNode.from_fields(
            result=IntegerValue,
        )
        add_node = AddNode(id="add1")

        workflow = Workflow(
            input_node=input_node,
            inner_nodes=[add_node],
            output_node=output_node,
            edges=[
                Edge.from_nodes(
                    source=input_node, source_key="a", target=add_node, target_key="a"
                ),
                Edge.from_nodes(
                    source=input_node, source_key="b", target=add_node, target_key="b"
                ),
                Edge.from_nodes(
                    source=add_node,
                    source_key="sum",
                    target=output_node,
                    target_key="result",
                ),
            ],
        )

        # Execute workflow
        context = InMemoryContext()
        input_data = {"a": IntegerValue(10), "b": IntegerValue(20)}

        result = await engine.execute(workflow, input_data, context)

        # Verify results
        assert result.status is WorkflowExecutionResultStatus.SUCCESS
        assert result.output["result"].root == 30

    @pytest.mark.asyncio
    async def test_execute_calls_load_internally(self):
        """Test that execute() calls load() internally before execution."""

        # Create engine
        engine = WorkflowEngine()

        # Create a simple workflow
        input_node = InputNode.from_fields(
            a=IntegerValue,
            b=IntegerValue,
        )
        output_node = OutputNode.from_fields(
            result=IntegerValue,
        )
        add_node = AddNode(id="add1")

        workflow = Workflow(
            input_node=input_node,
            inner_nodes=[add_node],
            output_node=output_node,
            edges=[
                Edge.from_nodes(
                    source=input_node, source_key="a", target=add_node, target_key="a"
                ),
                Edge.from_nodes(
                    source=input_node, source_key="b", target=add_node, target_key="b"
                ),
                Edge.from_nodes(
                    source=add_node,
                    source_key="sum",
                    target=output_node,
                    target_key="result",
                ),
            ],
        )

        # Execute workflow - load is called internally
        context = InMemoryContext()
        input_data = {"a": IntegerValue(7), "b": IntegerValue(4)}

        result = await engine.execute(workflow, input_data, context)

        # Verify it worked
        assert result.status is WorkflowExecutionResultStatus.SUCCESS
        assert result.output["result"].root == 11

    @pytest.mark.asyncio
    async def test_execute_with_typed_workflow(self):
        """Test that execute() works with already typed workflows."""

        # Create engine
        engine = WorkflowEngine()

        # Create workflow with typed node
        input_node = InputNode.from_fields(
            a=IntegerValue,
            b=IntegerValue,
        )
        output_node = OutputNode.from_fields(
            result=IntegerValue,
        )
        typed_node = AddNode(id="add1")

        workflow = Workflow(
            input_node=input_node,
            inner_nodes=[typed_node],
            output_node=output_node,
            edges=[
                Edge.from_nodes(
                    source=input_node, source_key="a", target=typed_node, target_key="a"
                ),
                Edge.from_nodes(
                    source=input_node, source_key="b", target=typed_node, target_key="b"
                ),
                Edge.from_nodes(
                    source=typed_node,
                    source_key="sum",
                    target=output_node,
                    target_key="result",
                ),
            ],
        )

        # Execute workflow
        context = InMemoryContext()
        input_data = {"a": IntegerValue(15), "b": IntegerValue(25)}

        result = await engine.execute(workflow, input_data, context)

        # Verify results
        assert result.status is WorkflowExecutionResultStatus.SUCCESS
        assert result.output["result"].root == 40

    @pytest.mark.asyncio
    async def test_execute_with_multiple_nodes(self):
        """Test execute() with a workflow containing multiple nodes."""

        # Create engine
        engine = WorkflowEngine()

        # Create workflow: (a + b) + c
        input_node = InputNode.from_fields(
            a=IntegerValue,
            b=IntegerValue,
            c=IntegerValue,
        )
        output_node = OutputNode.from_fields(
            result=IntegerValue,
        )
        add_node1 = AddNode(id="add1")
        add_node2 = AddNode(id="add2")

        workflow = Workflow(
            input_node=input_node,
            inner_nodes=[add_node1, add_node2],
            output_node=output_node,
            edges=[
                Edge.from_nodes(
                    source=input_node, source_key="a", target=add_node1, target_key="a"
                ),
                Edge.from_nodes(
                    source=input_node, source_key="b", target=add_node1, target_key="b"
                ),
                Edge.from_nodes(
                    source=add_node1, source_key="sum", target=add_node2, target_key="a"
                ),
                Edge.from_nodes(
                    source=input_node, source_key="c", target=add_node2, target_key="b"
                ),
                Edge.from_nodes(
                    source=add_node2,
                    source_key="sum",
                    target=output_node,
                    target_key="result",
                ),
            ],
        )

        # Execute: (2 + 3) + 4 = 9
        context = InMemoryContext()
        input_data = {
            "a": IntegerValue(2),
            "b": IntegerValue(3),
            "c": IntegerValue(4),
        }

        result = await engine.execute(workflow, input_data, context)

        # Verify results
        assert result.status is WorkflowExecutionResultStatus.SUCCESS
        assert result.output["result"].root == 9

    @pytest.mark.asyncio
    async def test_execute_with_tenant_specific_engine(self):
        """Test that tenant-specific engines execute with isolated registries."""

        # Create tenant-specific registry with only SampleAddNode
        tenant_registry = (
            NodeRegistry.builder(lazy=True)
            .register_node_class(Node)
            .register_node_class(SampleAddNode)
            .build()
        )

        engine = WorkflowEngine(
            node_registry=tenant_registry,
            value_registry=ValueRegistry.builder(lazy=True),
        )

        # Create workflow with SampleAdd node
        input_node = InputNode.empty()
        output_node = OutputNode.empty()

        workflow = Workflow.model_construct(
            input_node=input_node,
            inner_nodes=[
                Node.model_construct(type="SampleAdd", id="node1", version="1.0.0")
            ],
            output_node=output_node,
            edges=[],
        )

        # Execute workflow
        context = InMemoryContext()
        result = await engine.execute(workflow, {}, context)

        # Should execute successfully (no errors)
        assert result.status is WorkflowExecutionResultStatus.SUCCESS



# tests/test_node_migration.py
"""Integration tests for node migration during deserialization."""

import warnings
from collections.abc import Mapping
from functools import cached_property
from typing import Any, ClassVar, Literal

import pytest

from workflow_engine import NodeRegistry, StringValue, WorkflowEngine
from workflow_engine.core import Empty, Node, NodeTypeInfo, Params
from workflow_engine.core.migration import Migration, migration_registry
from workflow_engine.core.values import Data

# Test fixtures: a node type with migrations


class MigratableParams(Params):
    """Parameters for migratable test node."""

    value: StringValue


class MigratableOutput(Data):
    """Output for migratable test node."""

    result: StringValue


class MigratableNode(Node[Empty, MigratableOutput, MigratableParams]):
    """A node type used for testing migrations."""

    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="MigratableNode",
        display_name="Migratable Node",
        description="A test node for migration testing",
        version="2.0.0",  # Current version
        parameter_type=MigratableParams,
    )

    type: Literal["MigratableNode"] = "MigratableNode"  # pyright: ignore[reportIncompatibleVariableOverride]

    @cached_property
    def input_type(self):
        return Empty

    @cached_property
    def output_type(self):
        return MigratableOutput

    async def run(self, context: Any, input: Empty) -> MigratableOutput:
        return MigratableOutput(result=self.params.value)


class MigratableNodeMigration_1_0_0_to_2_0_0(Migration):
    """Migration from 1.0.0 to 2.0.0 for MigratableNode."""

    node_type = "MigratableNode"
    from_version = "1.0.0"
    to_version = "2.0.0"

    def migrate(self, data: Mapping[str, Any]) -> Mapping[str, Any]:
        result = dict(data)
        params = dict(result.get("params", {}))
        # In v1, the field was called 'text', in v2 it's 'value'
        if "text" in params:
            params["value"] = params.pop("text")
        result["params"] = params
        return result


@pytest.fixture(autouse=True)
def clean_migration_registry():
    """Clean the migration registry before and after each test."""
    # Store original migrations
    original_migrations = dict(migration_registry._migrations)

    yield

    # Restore original state
    migration_registry._migrations = original_migrations


class TestNodeMigration:
    """Integration tests for automatic node migration."""

    @pytest.mark.unit
    def test_load_current_version_no_migration(self):
        """Test that loading current version doesn't trigger migration."""
        data = {
            "type": "MigratableNode",
            "id": "test",
            "version": "2.0.0",
            "params": {"value": "hello"},
        }

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # First deserialize as base Node, then load via registry to get typed node
            untyped_node = Node.model_validate(data)
            node = NodeRegistry.DEFAULT.load_node(untyped_node)

            # Should not warn
            assert len(w) == 0

        assert isinstance(node, MigratableNode)
        assert node.version == "2.0.0"
        assert node.params.value.root == "hello"

    @pytest.mark.unit
    def test_load_old_version_with_migration(self):
        """Test that old version is automatically migrated."""
        # Register the migration
        migration_registry.register(MigratableNodeMigration_1_0_0_to_2_0_0)

        data = {
            "type": "MigratableNode",
            "id": "test",
            "version": "1.0.0",
            "params": {"text": "hello"},  # Old field name
        }

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # First deserialize as base Node, then load via registry to get typed node
            untyped_node = Node.model_validate(data)
            node = NodeRegistry.DEFAULT.load_node(untyped_node)

            # Should not warn since migration was successful
            assert len(w) == 0

        assert isinstance(node, MigratableNode)
        assert node.version == "2.0.0"  # Migrated to current version
        assert node.params.value.root == "hello"  # Field renamed

    @pytest.mark.unit
    def test_load_old_version_without_migration_warns(self):
        """Test that old version without migration issues warning."""
        # Don't register any migration

        data = {
            "type": "MigratableNode",
            "id": "test",
            "version": "1.5.0",  # Some old version with no migration
            "params": {"value": "hello"},
        }

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # First deserialize as base Node, then load via registry to get typed node
            untyped_node = Node.model_validate(data)
            node = NodeRegistry.DEFAULT.load_node(untyped_node)

            # Should have at least one warning about old version
            version_warnings = [
                warning
                for warning in w
                if "1.5.0 is older than" in str(warning.message)
            ]
            assert len(version_warnings) >= 1
            assert "may need to be migrated" in str(version_warnings[0].message)

        assert isinstance(node, MigratableNode)
        assert node.version == "1.5.0"  # Version unchanged (no migration)

    @pytest.mark.unit
    def test_migration_via_workflow_deserialization(self):
        """Test that migration works when loading workflow from JSON."""
        from workflow_engine import Workflow

        # Register the migration
        migration_registry.register(MigratableNodeMigration_1_0_0_to_2_0_0)

        workflow_data = {
            "input_node": {
                "type": "Input",
                "version": "1.0.0",
                "id": "input",
                "params": {"fields": {}},
            },
            "inner_nodes": [
                {
                    "type": "MigratableNode",
                    "id": "old_node",
                    "version": "1.0.0",
                    "params": {"text": "old_value"},
                }
            ],
            "output_node": {
                "type": "Output",
                "version": "1.0.0",
                "id": "output",
                "params": {"fields": {"output": {"type": "string"}}},
            },
            "edges": [
                {
                    "source_id": "old_node",
                    "source_key": "result",
                    "target_id": "output",
                    "target_key": "output",
                }
            ],
        }

        # Deserialize workflow, then load via engine to get typed nodes
        workflow = Workflow.model_validate(workflow_data)
        engine = WorkflowEngine()
        workflow = engine.load(workflow)

        # Node should be migrated (inner_nodes[0] is the MigratableNode)
        node = workflow.inner_nodes[0]
        assert isinstance(node, MigratableNode)
        assert node.version == "2.0.0"
        assert node.params.value.root == "old_value"

    @pytest.mark.unit
    def test_chained_migration(self):
        """Test that multi-step migrations work."""

        # Define intermediate migration
        class Migration_1_5_0_to_2_0_0(Migration):
            node_type = "MigratableNode"
            from_version = "1.5.0"
            to_version = "2.0.0"

            def migrate(self, data: Mapping[str, Any]) -> Mapping[str, Any]:
                result = dict(data)
                params = dict(result.get("params", {}))
                # In 1.5.0, field was 'content', in 2.0.0 it's 'value'
                if "content" in params:
                    params["value"] = params.pop("content")
                result["params"] = params
                return result

        class Migration_1_0_0_to_1_5_0(Migration):
            node_type = "MigratableNode"
            from_version = "1.0.0"
            to_version = "1.5.0"

            def migrate(self, data: Mapping[str, Any]) -> Mapping[str, Any]:
                result = dict(data)
                params = dict(result.get("params", {}))
                # In 1.0.0, field was 'text', in 1.5.0 it's 'content'
                if "text" in params:
                    params["content"] = params.pop("text")
                result["params"] = params
                return result

        migration_registry.register(Migration_1_0_0_to_1_5_0)
        migration_registry.register(Migration_1_5_0_to_2_0_0)

        data = {
            "type": "MigratableNode",
            "id": "test",
            "version": "1.0.0",
            "params": {"text": "hello"},
        }

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # First deserialize as base Node, then load via registry to get typed node
            untyped_node = Node.model_validate(data)
            node = NodeRegistry.DEFAULT.load_node(untyped_node)
            assert len(w) == 0

        assert node.version == "2.0.0"
        assert node.params.value.root == "hello"

    @pytest.mark.unit
    def test_migration_preserves_extra_fields(self):
        """Test that extra fields like position are preserved during migration."""
        migration_registry.register(MigratableNodeMigration_1_0_0_to_2_0_0)

        data = {
            "type": "MigratableNode",
            "id": "test",
            "version": "1.0.0",
            "params": {"text": "hello"},
            "position": {"x": 100, "y": 200},
        }

        # First deserialize as base Node, then load via registry to get typed node
        untyped_node = Node.model_validate(data)
        node = NodeRegistry.DEFAULT.load_node(untyped_node)

        # Extra field should be preserved
        dumped = node.model_dump()
        assert dumped["position"] == {"x": 100, "y": 200}

    @pytest.mark.unit
    def test_direct_class_validation_requires_current_schema(self):
        """Test that calling concrete class directly requires current version schema.

        Migration only works when loading via NodeRegistry.load_node(), not
        when calling the concrete class directly. This is by design - direct
        class validation bypasses the registry mechanism where migration occurs.

        Users should use NodeRegistry.load_node() for loading nodes with
        potentially old versions.
        """
        migration_registry.register(MigratableNodeMigration_1_0_0_to_2_0_0)

        # Old version data with old schema
        old_data = {
            "type": "MigratableNode",
            "id": "test",
            "version": "1.0.0",
            "params": {"text": "hello"},  # Old field name
        }

        # Direct validation fails because it doesn't go through migration
        from pydantic import ValidationError

        with pytest.raises(ValidationError):
            MigratableNode.model_validate(old_data)

        # But NodeRegistry.load_node works with migration
        untyped_node = Node.model_validate(old_data)
        node = NodeRegistry.DEFAULT.load_node(untyped_node)
        assert node.version == "2.0.0"
        assert node.params.value.root == "hello"

        # Direct validation works with current schema
        current_data = {
            "type": "MigratableNode",
            "id": "test",
            "version": "2.0.0",
            "params": {"value": "hello"},  # Current field name
        }
        node = MigratableNode.model_validate(current_data)
        assert node.version == "2.0.0"


class TestMigrationEdgeCases:
    """Test edge cases in migration."""

    @pytest.mark.unit
    def test_latest_version_not_migrated(self):
        """Test that 'latest' version is not migrated but resolved."""
        data = {
            "type": "MigratableNode",
            "id": "test",
            "version": "latest",
            "params": {"value": "hello"},
        }

        # First deserialize as base Node, then load via registry to get typed node
        untyped_node = Node.model_validate(data)
        node = NodeRegistry.DEFAULT.load_node(untyped_node)

        # Should resolve to current version
        assert node.version == "2.0.0"

    @pytest.mark.unit
    def test_newer_version_raises_error(self):
        """Test that newer version than current raises error."""
        from pydantic import ValidationError

        data = {
            "type": "MigratableNode",
            "id": "test",
            "version": "3.0.0",  # Newer than current 2.0.0
            "params": {"value": "hello"},
        }

        # First deserialize as base Node, then try to load via registry
        untyped_node = Node.model_validate(data)

        with pytest.raises(ValidationError) as exc_info:
            NodeRegistry.DEFAULT.load_node(untyped_node)

        assert "newer than the latest version" in str(exc_info.value)



# workflow_engine/core/migration/__init__.py
"""Node migration system for handling version upgrades."""

from .exceptions import (
    MigrationError,
    MigrationNotFoundError,
    MigrationValidationError,
)
from .migration import Migration
from .registry import MigrationRegistry, migration_registry
from .runner import MigrationRunner, migration_runner
from .workflow_migration import (
    clean_edges_after_migration,
    load_workflow_with_migration,
)

__all__ = [
    "Migration",
    "MigrationError",
    "MigrationNotFoundError",
    "MigrationRegistry",
    "MigrationRunner",
    "MigrationValidationError",
    "clean_edges_after_migration",
    "load_workflow_with_migration",
    "migration_registry",
    "migration_runner",
]



# tests/test_schema.py

"""
Tests two sets of functionalities:
1. that .to_value_cls() is the inverse of .to_value_schema()
2. that we can manually write JSON Schemas that will turn into the correct Value
   classes when .to_value_cls() is called on them.
3. that the aliasing system (e.g. { "x-value-type": "StringValue" } -> StringValue)
   works for all non-generic Value classes. { "title": "..." } also works as a
   backwards-compatible fallback.

In this file we follow the convention of using T for the expected type and U for
the type returned by .to_value_cls().
"""

import pytest

from workflow_engine import (
    BooleanValue,
    Data,
    Empty,
    FileValue,
    FloatValue,
    IntegerValue,
    NullValue,
    SequenceValue,
    StringMapValue,
    StringValue,
    ValueSchemaValue,
    WorkflowValue,
)
from workflow_engine.core.values import get_data_schema, validate_value_schema
from workflow_engine.core.values.schema import (
    BooleanValueSchema,
    BaseValueSchema,
    FieldSchemaMappingValue,
    FloatValueSchema,
    IntegerValueSchema,
    NullValueSchema,
    ReferenceValueSchema,
    SequenceValueSchema,
    StringMapValueSchema,
    StringValueSchema,
)
from workflow_engine.files import (
    JSONFileValue,
    JSONLinesFileValue,
    PDFFileValue,
    TextFileValue,
)

# ensure that these node types are registered for the workflow tests
from workflow_engine.nodes import (
    AddNode,  # noqa: F401
    ConstantIntegerNode,  # noqa: F401
)


@pytest.mark.unit
def test_boolean_schema_roundtrip():
    T = BooleanValue
    schema = T.to_value_schema()
    assert isinstance(schema, BooleanValueSchema)
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    t1 = T(True)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(False)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_boolean_schema_manual():
    T = BooleanValue
    json_schema = {
        "type": "boolean",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BooleanValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(False)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(True)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_boolean_schema_aliasing():
    T = BooleanValue
    json_schema = {
        "x-value-type": "BooleanValue",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BaseValueSchema)
    U = schema.to_value_cls()
    assert U == BooleanValue

    t1 = T(True)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(False)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_integer_schema_roundtrip():
    T = IntegerValue
    schema = T.to_value_schema()
    assert isinstance(schema, IntegerValueSchema)
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    t1 = T(42)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(2520)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_integer_schema_manual():
    T = IntegerValue
    json_schema = {
        "type": "integer",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, IntegerValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(-42)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(2520)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_integer_schema_with_extra_fields():
    """Constrained integer schemas produce subclasses that enforce the constraints."""
    json_schema = {
        "type": "integer",
        "minimum": 0,
        "maximum": 100,
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, IntegerValueSchema)
    U = schema.to_value_cls()

    assert issubclass(U, IntegerValue)
    assert U.model_json_schema()["minimum"] == 0
    assert U.model_json_schema()["maximum"] == 100

    assert U.model_validate(50).root == 50
    with pytest.raises(Exception):
        U.model_validate(-1)
    with pytest.raises(Exception):
        U.model_validate(101)


@pytest.mark.unit
def test_float_schema_with_extra_fields():
    """Constrained float schemas produce subclasses that enforce the constraints."""
    json_schema = {
        "type": "number",
        "minimum": 0.0,
        "maximum": 1.0,
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, FloatValueSchema)
    U = schema.to_value_cls()

    assert issubclass(U, FloatValue)
    assert U.model_json_schema()["minimum"] == 0.0
    assert U.model_json_schema()["maximum"] == 1.0

    assert U.model_validate(0.5).root == 0.5
    with pytest.raises(Exception):
        U.model_validate(-0.1)
    with pytest.raises(Exception):
        U.model_validate(1.1)


@pytest.mark.unit
def test_string_schema_with_extra_fields():
    """Constrained string schemas produce subclasses that enforce the constraints."""
    json_schema = {
        "type": "string",
        "minLength": 2,
        "maxLength": 5,
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, StringValueSchema)
    U = schema.to_value_cls()

    assert issubclass(U, StringValue)
    assert U.model_json_schema()["minLength"] == 2
    assert U.model_json_schema()["maxLength"] == 5

    assert U.model_validate("hi").root == "hi"
    with pytest.raises(Exception):
        U.model_validate("x")
    with pytest.raises(Exception):
        U.model_validate("toolong")


@pytest.mark.unit
def test_unknown_extras_preserved_in_constrained_cls():
    """Extras not in the field map are carried through as json_schema_extra."""
    json_schema = {
        "type": "number",
        "minimum": 0.0,
        "x-foo": "bar",
    }
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()

    assert issubclass(U, FloatValue)
    assert U.model_json_schema()["minimum"] == 0.0
    assert U.model_json_schema()["x-foo"] == "bar"
    with pytest.raises(Exception):
        U.model_validate(-0.1)


@pytest.mark.unit
def test_schema_from_pydantic_model_with_field_constraints():
    """Schemas extracted from Pydantic Field constraints produce enforcing subclasses."""
    from pydantic import BaseModel, Field

    class ModelWithConstraints(BaseModel):
        count: int = Field(ge=0, le=100)
        score: float = Field(ge=0.0, le=1.0)

    pydantic_schema = ModelWithConstraints.model_json_schema()
    count_schema = pydantic_schema["properties"]["count"]
    score_schema = pydantic_schema["properties"]["score"]

    count_cls = validate_value_schema(count_schema).to_value_cls()
    assert issubclass(count_cls, IntegerValue)
    count_cls.model_validate(50)
    with pytest.raises(Exception):
        count_cls.model_validate(101)

    score_cls = validate_value_schema(score_schema).to_value_cls()
    assert issubclass(score_cls, FloatValue)
    score_cls.model_validate(0.5)
    with pytest.raises(Exception):
        score_cls.model_validate(1.5)


@pytest.mark.unit
def test_integer_schema_aliasing():
    T = IntegerValue
    json_schema = {
        "x-value-type": "IntegerValue",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BaseValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(2048)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(2520)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_float_schema_roundtrip():
    T = FloatValue
    schema = T.to_value_schema()
    assert isinstance(schema, FloatValueSchema)
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    t1 = T(3.14159)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(2.71828)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_float_schema_manual():
    T = FloatValue
    json_schema = {
        "type": "number",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, FloatValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(2.71828)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(3.14159)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_float_schema_aliasing():
    T = FloatValue
    json_schema = {
        "x-value-type": "FloatValue",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BaseValueSchema)
    U = schema.to_value_cls()
    assert U == FloatValue

    t1 = T(1.41421)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(3.14159)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_null_schema_roundtrip():
    T = NullValue
    schema = T.to_value_schema()
    assert isinstance(schema, NullValueSchema)
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    t1 = T(None)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(None)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_null_schema_manual():
    T = NullValue
    json_schema = {
        "type": "null",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, NullValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(None)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(None)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_null_schema_aliasing():
    T = NullValue
    json_schema = {
        "x-value-type": "NullValue",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BaseValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(None)
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U(None)
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_string_schema_roundtrip():
    T = StringValue
    schema = T.to_value_schema()
    assert isinstance(schema, StringValueSchema)
    U = schema.to_value_cls()
    assert U == StringValue
    assert U.to_value_schema() == schema

    t1 = T("hello wengine")
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U("hi wengine")
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_string_schema_manual():
    T = StringValue
    json_schema = {
        "type": "string",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, StringValueSchema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T("salutations wengine")
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U("good morning wengine")
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_string_schema_aliasing():
    T = StringValue
    json_schema = {
        "x-value-type": "StringValue",
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, BaseValueSchema)
    U = schema.to_value_cls()
    assert U == StringValue

    t1 = T("hey wengine")
    u1 = U.model_validate(t1)
    assert u1 == t1

    u2 = U("good afternoon wengine")
    t2 = T.model_validate(u2)
    assert t2 == u2


@pytest.mark.unit
def test_sequence_schema_roundtrip():
    for ItemType, ItemSchema in (
        (BooleanValue, BooleanValueSchema),
        (FloatValue, FloatValueSchema),
        (IntegerValue, IntegerValueSchema),
        (NullValue, NullValueSchema),
        (StringValue, StringValueSchema),
    ):
        T = SequenceValue[ItemType]
        schema = T.to_value_schema()
        assert isinstance(schema, SequenceValueSchema)
        assert isinstance(schema.items, ReferenceValueSchema)
        assert isinstance(schema.defs[schema.items.id], ItemSchema)
        U = schema.to_value_cls()
        assert U == T
        assert U.to_value_schema() == schema


@pytest.mark.unit
def test_sequence_schema_manual():
    for type, ItemType, ItemSchema in (
        ("boolean", BooleanValue, BooleanValueSchema),
        ("number", FloatValue, FloatValueSchema),
        ("integer", IntegerValue, IntegerValueSchema),
        ("null", NullValue, NullValueSchema),
        ("string", StringValue, StringValueSchema),
    ):
        T = SequenceValue[ItemType]
        json_schema = {
            "type": "array",
            "items": {"type": type},
        }
        schema = validate_value_schema(json_schema)
        assert isinstance(schema, SequenceValueSchema)
        assert isinstance(schema.items, ItemSchema)
        U = schema.to_value_cls()
        assert U == T


@pytest.mark.unit
def test_sequence_schema_aliasing():
    for ItemType in (
        BooleanValue,
        FloatValue,
        IntegerValue,
        NullValue,
        StringValue,
    ):
        T = SequenceValue[ItemType]
        json_schema = {
            "type": "array",
            "items": {"x-value-type": ItemType.__name__},
        }
        schema = validate_value_schema(json_schema)
        assert isinstance(schema, SequenceValueSchema)
        assert isinstance(schema.items, BaseValueSchema)
        U = schema.to_value_cls()
        assert U == T


@pytest.mark.unit
def test_constrained_sequence_is_castable():
    """Constrained SequenceValue subclasses must remain compatible with the casting system."""
    import asyncio
    from workflow_engine.contexts import InMemoryContext

    json_schema = {
        "type": "array",
        "items": {"type": "number"},
        "minItems": 2,
        "maxItems": 4,
    }
    U = validate_value_schema(json_schema).to_value_cls()

    assert issubclass(U, SequenceValue)

    # Constraint is enforced at validation time
    assert U.model_validate([1.0, 2.0]) == [1.0, 2.0]
    with pytest.raises(Exception):
        U.model_validate([1.0])  # too short

    # A plain SequenceValue[FloatValue] can be cast to the constrained subclass
    source = SequenceValue[FloatValue].model_validate([1.0, 2.0, 3.0])
    context = InMemoryContext()
    result = asyncio.run(source.cast_to(U, context=context))
    assert isinstance(result, U)

    # A constrained instance can be cast to a plain SequenceValue[FloatValue]
    constrained = U.model_validate([1.0, 2.0])
    result2 = asyncio.run(
        constrained.cast_to(SequenceValue[FloatValue], context=context)
    )
    assert isinstance(result2, SequenceValue)


@pytest.mark.unit
def test_string_map_schema_roundtrip():
    for ItemType, ItemSchema in (
        (BooleanValue, BooleanValueSchema),
        (FloatValue, FloatValueSchema),
        (IntegerValue, IntegerValueSchema),
        (NullValue, NullValueSchema),
        (StringValue, StringValueSchema),
    ):
        T = StringMapValue[ItemType]
        schema = T.to_value_schema()
        assert isinstance(schema, StringMapValueSchema)
        assert isinstance(schema.additionalProperties, ReferenceValueSchema)
        assert isinstance(schema.defs[schema.additionalProperties.id], ItemSchema)
        U = schema.to_value_cls()
        assert U == T
        assert U.to_value_schema() == schema


@pytest.mark.unit
def test_string_map_schema_manual():
    for type, ItemType, ItemSchema in (
        ("boolean", BooleanValue, BooleanValueSchema),
        ("number", FloatValue, FloatValueSchema),
        ("integer", IntegerValue, IntegerValueSchema),
        ("null", NullValue, NullValueSchema),
        ("string", StringValue, StringValueSchema),
    ):
        T = StringMapValue[ItemType]
        json_schema = {
            "type": "object",
            "additionalProperties": {"type": type},
        }
        schema = validate_value_schema(json_schema)
        assert isinstance(schema, StringMapValueSchema)
        assert isinstance(schema.additionalProperties, ItemSchema)
        U = schema.to_value_cls()
        assert U == T


@pytest.mark.unit
def test_string_map_schema_aliasing():
    for ItemType in (
        BooleanValue,
        FloatValue,
        IntegerValue,
        NullValue,
        StringValue,
    ):
        T = StringMapValue[ItemType]
        json_schema = {
            "type": "object",
            "additionalProperties": {"x-value-type": ItemType.__name__},
        }
        schema = validate_value_schema(json_schema)
        assert isinstance(schema, StringMapValueSchema)
        assert isinstance(schema.additionalProperties, BaseValueSchema)
        U = schema.to_value_cls()
        assert U == T


@pytest.mark.unit
def test_super_recursive_schema_roundtrip():
    for T in (
        StringMapValue[SequenceValue[StringMapValue[StringValue]]],
        SequenceValue[StringMapValue[SequenceValue[NullValue]]],
        StringMapValue[StringMapValue[StringMapValue[IntegerValue]]],
        SequenceValue[SequenceValue[SequenceValue[BooleanValue]]],
    ):
        schema = T.to_value_schema()
        U = schema.to_value_cls()
        assert U == T
        assert U.to_value_schema() == schema


@pytest.mark.unit
def test_super_recursive_schema_manual():
    T = StringMapValue[SequenceValue[StringMapValue[StringValue]]]
    json_schema = {
        "type": "object",
        "additionalProperties": {
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": {"x-value-type": "StringValue"},
            },
        },
    }
    schema = validate_value_schema(json_schema)
    assert isinstance(schema, StringMapValueSchema)
    assert isinstance(schema.additionalProperties, SequenceValueSchema)
    assert isinstance(schema.additionalProperties.items, StringMapValueSchema)
    assert isinstance(
        schema.additionalProperties.items.additionalProperties, BaseValueSchema
    )
    U = schema.to_value_cls()
    assert U == T


@pytest.mark.unit
def test_empty_schema_roundtrip():
    T = Empty
    schema = get_data_schema(T)
    U = schema.to_value_cls()

    # for Empty, to_value_cls returns a new class not equal to the original
    # but it can serialize and deserialize instances of the original class
    t1 = T()
    u1 = U.model_validate(t1.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert u1.root.__dict__ == t1.__dict__

    u2 = U.model_validate({})
    t2 = T.model_validate(u2.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert t2.__dict__ == u2.root.__dict__


# defined outside of test_data_schema_roundtrip to get a proper class name
class FooBarData(Data):
    foo: StringValue
    bar: IntegerValue


@pytest.mark.unit
def test_data_schema_roundtrip():
    T = FooBarData
    schema = get_data_schema(T)
    U = schema.to_value_cls()

    # it can serialize and deserialize instances of the original class
    t1 = T(
        foo=StringValue("foo"),
        bar=IntegerValue(1),
    )
    u1 = U.model_validate(t1.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert u1.root.foo == t1.foo
    assert u1.root.bar == t1.bar
    assert u1.root.__dict__ == t1.__dict__

    u2 = U.model_validate(
        {
            "foo": "bar",
            "bar": 2,
        }
    )
    t2 = T.model_validate(u2.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert t2.foo == u2.root.foo
    assert t2.bar == u2.root.bar


@pytest.mark.unit
def test_data_schema_manual():
    T = FooBarData
    json_schema = {
        "title": "FooBarData",
        "type": "object",
        "properties": {
            "foo": {"type": "string"},
            "bar": {"type": "integer"},
        },
        "required": ["foo", "bar"],
    }
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()

    # it can serialize and deserialize instances of the original class
    t1 = T(
        foo=StringValue("bar"),
        bar=IntegerValue(12),
    )
    u1 = U.model_validate(t1.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert u1.root.foo == t1.foo
    assert u1.root.bar == t1.bar

    u2 = U.model_validate(
        {
            "foo": "baz",
            "bar": 24,
        }
    )
    t2 = T.model_validate(u2.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert u2.root.foo == t2.foo
    assert u2.root.bar == t2.bar


@pytest.mark.unit
def test_data_schema_aliasing():
    T = FooBarData
    json_schema = {
        "title": "FooBarData",
        "type": "object",
        "properties": {
            "foo": {"x-value-type": "StringValue"},
            "bar": {"x-value-type": "IntegerValue"},
        },
        "required": ["foo", "bar"],
    }
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()

    # it can serialize and deserialize instances of the original class
    t1 = T(
        foo=StringValue("foobar"),
        bar=IntegerValue(123),
    )
    u1 = U.model_validate(t1.model_dump())
    # equality check fails because they are technically different classes,
    # but they have the exact same fields
    assert u1.root.foo == t1.foo
    assert u1.root.bar == t1.bar

    u2 = U.model_validate(
        {
            "foo": "baz",
            "bar": 24,
        }
    )
    t2 = T.model_validate(u2.model_dump())
    assert u2.root.foo == t2.foo
    assert u2.root.bar == t2.bar


@pytest.mark.unit
def test_data_schema_with_multiple_constrained_float_properties():
    """Two FloatValue properties with different constraints must not collide in $defs."""
    json_schema = {
        "title": "BoundedData",
        "type": "object",
        "properties": {
            "probability": {"type": "number", "minimum": 0, "maximum": 1},
            "temperature": {"type": "number", "minimum": -273.15},
        },
        "required": ["probability", "temperature"],
    }
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()

    # Both fields enforce their own distinct constraints
    instance = U.model_validate({"probability": 0.5, "temperature": 20.0})
    assert instance.root.probability.root == 0.5
    assert instance.root.temperature.root == 20.0

    with pytest.raises(Exception):
        U.model_validate({"probability": 1.5, "temperature": 20.0})

    with pytest.raises(Exception):
        U.model_validate({"probability": 0.5, "temperature": -300.0})

    # Schema must have two distinct $defs entries, not one clobbering the other
    full_schema = U.model_json_schema()
    defs = full_schema.get("$defs", {})
    float_defs = [v for v in defs.values() if v.get("type") == "number"]
    assert len(float_defs) == 2


@pytest.mark.unit
def test_field_schema_mapping_preserves_extras():
    """FieldSchemaMappingValue preserves schema extras (both known and unknown) on property schemas."""
    from workflow_engine.core.values.schema import ValueSchemaValue

    score_schema = FloatValueSchema(
        **{"type": "number", "minimum": 0.0, "maximum": 1.0, "x-foo": "bar"}
    )
    mapping = FieldSchemaMappingValue({"score": ValueSchemaValue(score_schema)})
    data_schema = mapping.to_data_schema("TestData")
    U = data_schema.to_value_cls()

    # Known constraints are enforced
    assert U.model_validate({"score": 0.5}).root.score.root == 0.5
    with pytest.raises(Exception):
        U.model_validate({"score": 1.5})
    with pytest.raises(Exception):
        U.model_validate({"score": -0.1})

    # Unknown extras survive into the property's schema
    score_field_schema = U.model_json_schema()["$defs"]
    score_def = next(
        v for v in score_field_schema.values() if v.get("type") == "number"
    )
    assert score_def["x-foo"] == "bar"


@pytest.mark.unit
def test_file_schema_roundtrip():
    for T in (
        FileValue,
        JSONFileValue,
        JSONLinesFileValue,
        PDFFileValue,
        TextFileValue,
    ):
        schema = T.to_value_schema()
        U = schema.to_value_cls()
        assert U == T
        assert U.to_value_schema() == schema

        t1 = T.from_path("foo", foo="bar", bar="baz")
        u1 = U.model_validate(t1.model_dump())
        assert u1 == t1

        u2 = U.model_validate({"path": "bar", "metadata": {"baz": 3}})
        t2 = T.model_validate(u2.model_dump())
        assert t2.path == u2.root.path
        assert t2.metadata == u2.root.metadata


@pytest.mark.unit
def test_file_schema_aliasing():
    for T in (
        FileValue,
        JSONFileValue,
        JSONLinesFileValue,
        PDFFileValue,
        TextFileValue,
    ):
        json_schema = {"x-value-type": T.__name__}
        schema = validate_value_schema(json_schema)
        U = schema.to_value_cls()
        assert U == T

        t1 = T.from_path("bar", bar="baz", baz="foo")
        u1 = U.model_validate(t1.model_dump())
        assert u1.root.path == t1.path
        assert u1.root.metadata == t1.metadata

        u2 = U.model_validate({"path": "bar", "metadata": {"baz": 3}})
        t2 = T.model_validate(u2.model_dump())
        assert t2.path == u2.root.path
        assert t2.metadata == u2.root.metadata


@pytest.mark.unit
def test_workflow_schema_roundtrip():
    T = WorkflowValue
    schema = T.to_value_schema()
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    with open("examples/addition.json", "r") as f:
        workflow_json = f.read().strip()
    t1 = T.model_validate_json(workflow_json)
    u1 = U.model_validate_json(workflow_json)
    assert u1 == t1


@pytest.mark.unit
def test_workflow_schema_aliasing():
    T = WorkflowValue
    json_schema = {"x-value-type": T.__name__}
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()
    assert U == T

    with open("examples/addition.json", "r") as f:
        workflow_json = f.read().strip()

    t1 = T.model_validate_json(workflow_json)
    u1 = U.model_validate_json(workflow_json)
    assert u1 == t1


@pytest.mark.unit
def test_value_schema_value_roundtrip():
    T = ValueSchemaValue
    schema = T.to_value_schema()
    U = schema.to_value_cls()
    assert U == T
    assert U.to_value_schema() == schema

    t1 = T(StringValueSchema(type="string", title="StringValue"))
    u1 = U.model_validate(t1.model_dump())
    assert u1 == t1

    u2 = U(IntegerValueSchema(type="integer", title="IntegerValue"))
    t2 = T.model_validate(u2.model_dump())
    assert t2 == u2


@pytest.mark.unit
def test_value_schema_value_manual():
    T = ValueSchemaValue
    json_schema = {"x-value-type": T.__name__}
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(BooleanValueSchema(type="boolean", title="BooleanValue"))
    u1 = U.model_validate(t1.model_dump())
    assert u1 == t1

    u2 = U(NullValueSchema(type="null", title="NullValue"))
    t2 = T.model_validate(u2.model_dump())
    assert t2 == u2


@pytest.mark.unit
def test_value_schema_value_aliasing():
    T = ValueSchemaValue
    json_schema = {"x-value-type": T.__name__}
    schema = validate_value_schema(json_schema)
    U = schema.to_value_cls()
    assert U == T

    t1 = T(FloatValueSchema(type="number", title="FloatValue"))
    u1 = U.model_validate(t1.model_dump())
    assert u1 == t1

    u2 = U(
        SequenceValueSchema(
            type="array",
            items=StringValueSchema(type="string", title="StringValue"),
        )
    )
    t2 = T.model_validate(u2.model_dump())
    assert t2 == u2



# workflow_engine/core/values/value.py
from __future__ import annotations

import inspect
import re

from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping
from functools import cached_property
from hashlib import md5
from logging import getLogger
from typing import (
    TYPE_CHECKING,
    Awaitable,
    ClassVar,
    Generic,
    Literal,
    Protocol,
    Self,
    TypeVar,
    get_origin,
    overload,
)

from overrides import override
from pydantic import PrivateAttr

from ...utils.immutable import ImmutableRootModel

if TYPE_CHECKING:
    from ..context import Context
    from .schema import ValueSchema


logger = getLogger(__name__)


T = TypeVar("T")
V = TypeVar("V", bound="Value")
type ValueType = type[Value]


def get_origin_and_args(t: ValueType) -> tuple[ValueType, tuple[ValueType, ...]]:
    """
    For a non-generic value type NonGenericValue, returns (NonGenericValue, ()).

    For a generic value type GenericValue[Argument1Value, Argument2Value, ...],
    returns (GenericValue, (Argument1Value, Argument2Value, ...)).
    All arguments must themselves be Value subclasses.
    """
    # Pydantic RootModels don't play nice with get_origin and get_args, so we
    # get the root type directly from the model fields.
    assert issubclass(t, Value)
    info = t.__pydantic_generic_metadata__
    origin = info["origin"]
    args = info["args"]
    if origin is None:
        assert len(args) == 0
        return t, ()
    else:
        assert issubclass(origin, Value)
        assert len(args) > 0
        return origin, tuple(args)


type ValueTypeKey = tuple[str, tuple[ValueTypeKey, ...]]


def get_value_type_key(t: ValueType) -> ValueTypeKey:
    """
    Get a unique hashable key for a Value type.
    If t is a generic type, recursively call `get_value_type_key` to expand any
    Value types in the args.
    """
    origin, args = get_origin_and_args(t)
    return origin.__name__, tuple(
        get_value_type_key(arg) if issubclass(arg, Value) else arg for arg in args
    )


SourceType = TypeVar("SourceType", bound="Value")
TargetType = TypeVar("TargetType", bound="Value")


class Caster(Protocol, Generic[SourceType, TargetType]):  # type: ignore
    """
    A caster is a contextual function that transforms the type of a Value.

    May be either a sync or async function.
    """

    def __call__(
        self,
        value: SourceType,
        context: "Context",
    ) -> TargetType | Awaitable[TargetType]: ...


class GenericCaster(Protocol, Generic[SourceType, TargetType]):  # type: ignore
    """
    A generic caster is a contextual function that takes a source type and a
    target type, and outputs a caster between the two types, or None if the cast
    is not possible.
    This is an advanced feature intended for use on generic types.

    The purpose of this two-step approach is to explicitly allow or deny type
    casts before the exact type of the value is known. This is necessary because
    the type of a Value is not known until the Value is created.
    """

    def __call__(
        self,
        source_type: type[SourceType],
        target_type: type[TargetType],
    ) -> Caster[SourceType, TargetType] | None: ...


generic_pattern = re.compile(r"^[a-zA-Z]\w+\[.*\]$")


class Value(ImmutableRootModel[T], Generic[T]):
    """
    Wraps an arbitrary read-only value which can be passed as input to a node.

    Each Value subclass defines a specific type (possibly generic) of value.
    After defining the subclass, you can register Caster functions to convert
    other Value classes to that type, using the register_cast_to decorator.
    Casts are registered in any order.

    Each Value subclass inherits its parent classes' Casters.
    To avoid expanding the type tree every time, we cache the Casters at each
    class the first time a cast is used.
    Once that cache is created, the casts are locked and can no longer be
    changed.
    """

    # these properties force us to implement __eq__ and __hash__ to ignore them
    _casters: ClassVar[dict[str, GenericCaster]] = {}
    _resolved_casters: ClassVar[dict[str, GenericCaster] | None] = None
    _cast_cache: dict[ValueTypeKey, "Value"] = PrivateAttr(
        default_factory=dict,
    )

    def __init_subclass__(cls, register: bool = True, **kwargs):
        super().__init_subclass__(**kwargs)

        # reinitialize for each subclass so it doesn't just reference the parent
        cls._casters = {}
        cls._resolved_casters = None

        if not register:
            return

        while generic_pattern.match(cls.__name__) is not None:
            assert cls.__base__ is not None
            cls = cls.__base__
        # Skip generic base classes (e.g. SequenceValue, StringMapValue) whose
        # __pydantic_generic_metadata__["parameters"] still contains unbound
        # TypeVars — only fully-concrete classes belong in the registry.
        if (
            get_origin(cls) is None
            and len(cls.__pydantic_generic_metadata__["parameters"]) == 0
        ):
            ValueRegistry.DEFAULT.register_value_class(cls)

    @classmethod
    def _get_casters(cls) -> dict[str, GenericCaster]:
        """
        Get all type casting functions for this class, including those inherited
        from parent classes.
        This inherits from all parents classes, though they will be overridden
        if the child class has its own casting function for the same type.
        """
        if cls._resolved_casters is not None:
            return cls._resolved_casters

        resolved_casters: dict[str, GenericCaster] = cls._casters.copy()

        # Add converters from all classes in MRO order
        # (starting from this class, then parents, then parents of parents, ...)
        for parent in cls.__bases__:
            if issubclass(parent, Value):
                parent_casters = parent._get_casters()
            else:
                continue

            for origin, caster in parent_casters.items():
                # converters in the child class override those in the parent
                # class
                if origin not in resolved_casters:
                    resolved_casters[origin] = caster

        cls._resolved_casters = resolved_casters
        return resolved_casters

    @classmethod
    def register_cast_to(cls, t: type[V]):
        """
        A decorator to register a possible type cast from this class to the
        class T, neither of which are generic.
        """

        def wrap(caster: Caster[Self, V]):
            cls.register_generic_cast_to(t)(lambda source_type, target_type: caster)
            return caster

        return wrap

    @classmethod
    def register_generic_cast_to(cls, t: type[V]):
        """
        A decorator to register a possible type cast from this class to the
        class T, either of which may be generic.
        """

        def wrap(caster: GenericCaster[Self, V]):
            if cls._resolved_casters is not None:
                raise RuntimeError(
                    f"Cannot add casters for {cls.__name__} after it has been used to cast values"
                )

            target_origin, _ = get_origin_and_args(t)
            name = target_origin.__name__
            if name in cls._casters:
                raise AssertionError(
                    f"Type caster from {cls.__name__} to {name} already registered"
                )
            cls._casters[name] = caster

        return wrap

    @classmethod
    def get_caster(cls, t: type[V]) -> Caster[Self, V] | None:
        converters = cls._get_casters()
        target_origin, _ = get_origin_and_args(t)
        if target_origin.__name__ in converters:
            generic_caster = converters[target_origin.__name__]
            caster = generic_caster(cls, t)
            if caster is not None:
                return caster

        if issubclass(cls, t):
            return lambda value, context: value  # type: ignore

        return None

    @classmethod
    def can_cast_to(cls, t: type[V]) -> bool:
        """
        Returns True if there is any hope of casting this value to the type t.
        """
        return cls.get_caster(t) is not None

    def __eq__(self, other) -> bool:
        if isinstance(other, Value):
            return self.root == other.root
        return self.root == other

    def __hash__(self):
        return hash(self.root)

    def __str__(self) -> str:
        return str(self.root)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.root})"

    def __bool__(self) -> bool:
        return bool(self.root)

    @cached_property
    def md5(self) -> str:
        return md5(str(self).encode()).hexdigest()

    async def cast_to(self, t: type[V], *, context: "Context") -> V:
        key = get_value_type_key(t)
        if key in self._cast_cache:
            casted: V = self._cast_cache[key]  # type: ignore
            return casted

        cast_fn = self.__class__.get_caster(t)
        if cast_fn is not None:
            result = cast_fn(self, context)
            casted: V = (await result) if inspect.iscoroutine(result) else result  # type: ignore
            self._cast_cache[key] = casted
            return casted

        raise ValueError(f"Cannot convert {self} to {t}")

    @classmethod
    async def cast_from(cls, v: "Value", *, context: "Context") -> Self:
        return await v.cast_to(cls, context=context)

    @classmethod
    def to_value_schema(cls) -> "ValueSchema":
        from .schema import validate_value_schema  # avoid circular import

        schema = cls.model_json_schema()
        schema["x-value-type"] = cls.__name__
        return validate_value_schema(schema)


class ValueRegistry(ABC):
    """
    An immutable registry of value types by name.
    """

    DEFAULT: ClassVar[LazyValueRegistry]

    @abstractmethod
    def get_value_class(self, name: str) -> ValueType:
        """Get a value type by name."""
        pass

    @abstractmethod
    def has_name(self, name: str) -> bool:
        """Check if a value type name is registered."""
        pass

    @abstractmethod
    def all_value_classes(self) -> Iterable[tuple[str, ValueType]]:
        """Return all registered value classes as (name, class) pairs."""
        pass

    def load_value(self, schema: ValueSchema) -> ValueType | None:
        """
        Load a value type from a schema by looking up the registry.

        Checks x-value-type first, then falls back to title for backwards
        compatibility. If no match is found, returns None to indicate that the
        caller should fall back to building the value class from the schema.

        Args:
            schema: A ValueSchema with an optional value_type or title field

        Returns:
            The registered value class if a match is found, None otherwise
        """
        if schema.value_type is not None and self.has_name(schema.value_type):
            return self.get_value_class(schema.value_type)
        if schema.title is not None and self.has_name(schema.title):
            return self.get_value_class(schema.title)
        return None

    @overload
    @staticmethod
    def builder(*, lazy: Literal[True]) -> LazyValueRegistry: ...

    @overload
    @staticmethod
    def builder(*, lazy: Literal[False] = False) -> EagerValueRegistryBuilder: ...

    @staticmethod
    def builder(*, lazy: bool = False):
        return LazyValueRegistry() if lazy else EagerValueRegistryBuilder()

    @overload
    def extend(self, *, lazy: Literal[True]) -> LazyValueRegistry: ...

    @overload
    def extend(self, *, lazy: Literal[False] = False) -> EagerValueRegistryBuilder: ...

    def extend(self, *, lazy: bool = False):
        """Create a new builder pre-populated with all value classes from this registry."""
        builder = ValueRegistry.builder(lazy=lazy)
        for _name, value_cls in self.all_value_classes():
            builder.register_value_class(value_cls)
        return builder


class ValueRegistryBuilder(ABC):
    """
    A builder for creating value registries.
    """

    @abstractmethod
    def register_value_class(self, value_cls: ValueType) -> Self:
        """Register a value type using its class name."""
        pass

    @abstractmethod
    def remove_value_class(
        self, value_cls: ValueType, *, missing_ok: bool = False
    ) -> Self:
        """Remove a value type by class."""
        pass

    @abstractmethod
    def build(self) -> ValueRegistry:
        """Build and return the registry."""
        pass


class ImmutableValueRegistry(ValueRegistry):
    """
    An ImmutableValueRegistry is a ValueRegistry that is immutable after
    construction; enforced by shallow copying the input data structures.
    """

    def __init__(self, *, value_classes: Mapping[str, ValueType]):
        self._value_classes = dict(value_classes)

    @override
    def has_name(self, name: str) -> bool:
        return name in self._value_classes

    @override
    def get_value_class(self, name: str) -> ValueType:
        if name not in self._value_classes:
            raise ValueError(f'Value type "{name}" is not registered')
        return self._value_classes[name]

    @override
    def all_value_classes(self) -> Iterable[tuple[str, ValueType]]:
        return self._value_classes.items()


class EagerValueRegistryBuilder(ValueRegistryBuilder):
    """
    A builder that validates registrations immediately.
    """

    def __init__(self):
        self._value_classes: dict[str, ValueType] = {}

    @override
    def register_value_class(self, value_cls: ValueType) -> Self:
        name = value_cls.__name__
        if name in self._value_classes:
            conflict = self._value_classes[name]
            if value_cls is not conflict:
                raise ValueError(
                    f'Value type "{name}" (class {value_cls.__name__}) is already '
                    f"registered to a different class ({conflict.__name__})"
                )
        self._value_classes[name] = value_cls
        logger.debug("Registering value type %s", name)
        return self

    @override
    def remove_value_class(
        self, value_cls: ValueType, *, missing_ok: bool = False
    ) -> Self:
        name = value_cls.__name__
        if name not in self._value_classes:
            if not missing_ok:
                raise ValueError(f'Value type "{name}" is not registered')
        else:
            del self._value_classes[name]
        return self

    @override
    def build(self) -> ValueRegistry:
        return ImmutableValueRegistry(value_classes=self._value_classes)


class LazyValueRegistry(ValueRegistry, ValueRegistryBuilder):
    """
    A LazyValueRegistry is both a ValueRegistry and a ValueRegistryBuilder with a
    2-part lifecycle:
    1. Registration phase: acts as a ValueRegistryBuilder until .build() is called.
    2. Registry phase: acts as a ValueRegistry.

    Validations are deferred until the registry is frozen; this allows for a
    LazyValueRegistry to exist as long as it is not used, making it suitable for
    automatic value registration.
    """

    def __init__(self):
        self._registrations: list[ValueType] = []
        self._removals: dict[ValueType, bool] = {}  # cls -> missing_ok
        self._frozen: bool = False

        # initialized after freeze
        self._value_classes: Mapping[str, ValueType]

    # REGISTRATION PHASE

    @override
    def register_value_class(self, value_cls: ValueType) -> Self:
        if self._frozen:
            # Allow re-registration of the same class (for testing/reloading)
            name = value_cls.__name__
            if name in self._value_classes and self._value_classes[name] is value_cls:
                logger.debug(
                    "Value type %s already registered, skipping",
                    name,
                )
                return self
            raise ValueError(
                f"Value registry is frozen, cannot register new value type '{name}' (class {value_cls.__name__})"
            )
        self._registrations.append(value_cls)
        return self

    @override
    def remove_value_class(
        self, value_cls: ValueType, *, missing_ok: bool = False
    ) -> Self:
        if self._frozen:
            raise ValueError("Value registry is frozen, cannot remove value types.")
        self._removals[value_cls] = missing_ok
        return self

    @override
    def build(self) -> ValueRegistry:
        if self._frozen:
            return self

        self._frozen = True
        _value_classes = {}
        for value_cls in self._registrations:
            name = value_cls.__name__
            if name in _value_classes:
                conflict = _value_classes[name]
                if value_cls is conflict:
                    logger.warning(
                        "Value type %s is already registered to class %s, skipping registration",
                        name,
                        value_cls.__name__,
                    )
                else:
                    raise ValueError(
                        f'Value type "{name}" (class {value_cls.__name__}) is already '
                        f"registered to a different class ({conflict.__name__})"
                    )
            else:
                _value_classes[name] = value_cls
                logger.debug("Registering value type %s", name)

        for value_cls, missing_ok in self._removals.items():
            name = value_cls.__name__
            if name not in _value_classes:
                if not missing_ok:
                    raise ValueError(f'Value type "{name}" is not registered')
            else:
                del _value_classes[name]

        del self._registrations  # Memory optimization
        del self._removals
        self._value_classes = _value_classes

        return self

    # USE PHASE

    @override
    def has_name(self, name: str) -> bool:
        self.build()
        return name in self._value_classes

    @override
    def get_value_class(self, name: str) -> ValueType:
        self.build()
        if name not in self._value_classes:
            raise ValueError(f'Value type "{name}" is not registered')
        return self._value_classes[name]

    @override
    def all_value_classes(self) -> Iterable[tuple[str, ValueType]]:
        self.build()
        return self._value_classes.items()


ValueRegistry.DEFAULT = ValueRegistry.builder(lazy=True)


__all__ = [
    "Caster",
    "GenericCaster",
    "get_origin_and_args",
    "Value",
    "ValueRegistry",
    "ValueType",
]



# workflow_engine/nodes/iteration.py
"""
Nodes that iterate over a sequence of items.
"""

from functools import cached_property
from typing import ClassVar, Literal, Self

from overrides import override
from pydantic import Field

from workflow_engine.core.io import SchemaParams

from ..core import (
    Context,
    DataValue,
    Edge,
    Empty,
    InputNode,
    Node,
    NodeTypeInfo,
    OutputNode,
    Params,
    SequenceValue,
    Workflow,
    WorkflowValue,
)
from ..core.values.data import get_field_annotations, get_only_field
from .data import (
    ExpandDataNode,
    ExpandSequenceNode,
    GatherDataNode,
    GatherSequenceNode,
    SequenceData,
)


class ForEachParams(Params):
    workflow: WorkflowValue = Field(
        title="Workflow", description="The workflow to run for each item."
    )


class ForEachNode(Node[SequenceData, SequenceData | Empty, ForEachParams]):
    """
    A node that executes the internal workflow W for each item in the input
    sequence.

    For each item i in the input sequence, create a copy of W, call it W[i].
    The implementation automatically chooses the most efficient path:

    - **Single input field**: Sequence of scalar values, direct wiring from
      expand to each workflow (no ExpandDataNode).
    - **Multiple input fields**: Sequence of Data objects, ExpandDataNode
      adapter per item.
    - **Single output field**: Direct wiring from each workflow to gather
      (no GatherDataNode).
    - **Multiple output fields**: GatherDataNode adapter per item.
    - **Zero output fields**: Outputs nothing (Empty); inner workflows run for
      side effects only.
    """

    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="ForEach",
        display_name="ForEach",
        description="Executes the internal workflow for each item in the input sequence.",
        version="1.0.0",
        parameter_type=ForEachParams,
    )

    type: Literal["ForEach"] = "ForEach"  # pyright: ignore[reportIncompatibleVariableOverride]

    @cached_property
    def workflow(self) -> Workflow:
        return self.params.workflow.root

    @cached_property
    def _input_field_count(self) -> int:
        return len(get_field_annotations(self.workflow.input_type))

    @cached_property
    def _output_field_count(self) -> int:
        return len(get_field_annotations(self.workflow.output_type))

    def _has_single_input(self) -> bool:
        return self._input_field_count == 1

    def _has_single_output(self) -> bool:
        return self._output_field_count == 1

    def _has_no_output(self) -> bool:
        return self._output_field_count == 0

    @cached_property
    def _input_item_type(self):
        """Single input field type; only valid when _has_single_input()."""
        return get_only_field(self.workflow.input_type)[1]

    @cached_property
    def _output_item_type(self):
        """Single output field type; only valid when _has_single_output()."""
        return get_only_field(self.workflow.output_type)[1]

    def _input_element_type(self):
        if self._has_single_input():
            return self._input_item_type
        return DataValue[self.workflow.input_type]

    def _output_element_type(self):
        """Only valid when _has_no_output() is False."""
        assert not self._has_no_output()
        if self._has_single_output():
            return self._output_item_type
        return DataValue[self.workflow.output_type]

    @cached_property
    def input_type(self):
        return SequenceData[self._input_element_type()]

    @cached_property
    def output_type(self):
        if self._has_no_output():
            return Empty
        return SequenceData[self._output_element_type()]

    def _build_input_output_nodes(self, n: int) -> tuple[InputNode, OutputNode]:
        input_seq_type = SequenceValue[self._input_element_type()]
        input_params = SchemaParams.from_fields(sequence=input_seq_type)
        input_node = InputNode(id="input", params=input_params)
        if self._has_no_output():
            output_node = OutputNode.empty()
        else:
            output_seq_type = SequenceValue[self._output_element_type()]
            output_params = SchemaParams.from_fields(sequence=output_seq_type)
            output_node = OutputNode(id="output", params=output_params)
        return input_node, output_node

    @override
    async def run(self, context: Context, input: SequenceData) -> Workflow:
        n = len(input.sequence)
        input_node, output_node = self._build_input_output_nodes(n)
        has_no_output = self._has_no_output()

        expand_element_type = self._input_element_type()
        expand = ExpandSequenceNode.from_length(
            id="expand",
            length=n,
            element_type=expand_element_type,
        )

        inner_nodes: list[Node] = [expand]
        edges: list[Edge] = [
            Edge.from_nodes(
                source=input_node,
                source_key="sequence",
                target=expand,
                target_key="sequence",
            ),
        ]

        gather: GatherSequenceNode | None = None
        if not has_no_output:
            gather = GatherSequenceNode.from_length(
                id="gather",
                length=n,
                element_type=self._output_element_type(),
            )
            inner_nodes.append(gather)
            edges.append(
                Edge.from_nodes(
                    source=gather,
                    source_key="sequence",
                    target=output_node,
                    target_key="sequence",
                )
            )

        for i in range(n):
            namespace = f"element_{i}"
            item_workflow = self.workflow.with_namespace(namespace)

            input_adapter = (
                ExpandDataNode.from_data_type(
                    id="input_adapter",
                    data_type=self.workflow.input_type,
                ).with_namespace(namespace)
                if not self._has_single_input()
                else None
            )
            output_adapter = (
                GatherDataNode.from_data_type(
                    id="output_adapter",
                    data_type=self.workflow.output_type,
                ).with_namespace(namespace)
                if self._has_single_output() is False and has_no_output is False
                else None
            )

            if input_adapter is not None:
                inner_nodes.append(input_adapter)
            inner_nodes.extend(item_workflow.inner_nodes)
            if output_adapter is not None:
                inner_nodes.append(output_adapter)

            if input_adapter is not None:
                edges.append(
                    Edge.from_nodes(
                        source=expand,
                        source_key=expand.key(i),
                        target=input_adapter,
                        target_key="data",
                    )
                )

            for edge in item_workflow.edges:
                if edge.source_id == item_workflow.input_node.id:
                    source_id = (
                        input_adapter.id if input_adapter is not None else expand.id
                    )
                    source_key = (
                        edge.source_key if input_adapter is not None else expand.key(i)
                    )
                else:
                    source_id = edge.source_id
                    source_key = edge.source_key
                if edge.target_id == item_workflow.output_node.id:
                    if has_no_output:
                        continue
                    assert gather is not None
                    target_id = (
                        output_adapter.id if output_adapter is not None else gather.id
                    )
                    target_key = (
                        edge.target_key if output_adapter is not None else gather.key(i)
                    )
                else:
                    target_id = edge.target_id
                    target_key = edge.target_key
                edges.append(
                    Edge(
                        source_id=source_id,
                        source_key=source_key,
                        target_id=target_id,
                        target_key=target_key,
                    )
                )

            if output_adapter is not None:
                assert gather is not None
                edges.append(
                    Edge.from_nodes(
                        source=output_adapter,
                        source_key="data",
                        target=gather,
                        target_key=gather.key(i),
                    )
                )

        return Workflow(
            input_node=input_node,
            inner_nodes=inner_nodes,
            output_node=output_node,
            edges=edges,
        )

    @classmethod
    def from_workflow(cls, id: str, workflow: Workflow) -> Self:
        return cls(id=id, params=ForEachParams(workflow=WorkflowValue(workflow)))


__all__ = [
    "ForEachNode",
]



from collections.abc import MutableMapping
from typing import Any, Generic, Self, TypeVar
from pydantic import BaseModel, ConfigDict, RootModel


_immutable_model_config = ConfigDict(
    frozen=True,
    validate_assignment=True,
)


class _ImmutableMixin:
    """
    A base model that is immutable.
    """

    def __init_subclass__(cls, **kwargs):
        """
        Add the immutable model config to the rest of this class' model config.
        """
        cls.model_config = cls.model_config | _immutable_model_config
        super().__init_subclass__(**kwargs)


class ImmutableBaseModel(BaseModel, _ImmutableMixin):
    def model_update(self, **kwargs: Any) -> Self:
        """
        Returns a new copy of the model with the given kwargs updated, and
        validates the assignment.
        Contrary to common belief, setting revalidate_instances="always" and
        validate_assignment=True is not enough, because model_copy bypasses all
        of that.
        See https://github.com/pydantic/pydantic/issues/418
        """
        assert isinstance(self, BaseModel)
        updated = self.model_copy(update=kwargs)  # type: ignore
        self.__class__.model_validate(updated.__dict__)  # validate the updated model

        # Clear cached properties that may have been copied from the original instance
        # model_copy copies __dict__ which includes functools.cached_property values
        model_field_names = set(self.__class__.model_fields.keys())
        cached_keys = [k for k in updated.__dict__ if k not in model_field_names]
        updated_dict = updated.__dict__
        assert isinstance(updated_dict, MutableMapping)
        for key in cached_keys:
            del updated_dict[key]

        return updated

    def _model_mutate(self, **kwargs: Any):
        """
        Mutates the model in place despite its immutability.
        This is obviously dangerous, so the method is protected.
        To make it slightly safer, we validate the assignment via model_update.
        """
        updated = self.model_update(**kwargs)
        assert isinstance(self.__dict__, MutableMapping)
        self.__dict__.update(updated.__dict__)


T = TypeVar("T")


class ImmutableRootModel(RootModel[T], _ImmutableMixin, Generic[T]):
    pass


__all__ = [
    "ImmutableBaseModel",
    "ImmutableRootModel",
]



import pytest

from workflow_engine.utils.hash import json_digest


@pytest.mark.unit
def test_json_digest_stable():
    """Hash must not change across refactors — if this fails, digest values in
    schema.py will silently differ from ones produced by older code."""
    data = {"name": "test", "values": [1, 2, 3], "meta": {"active": True}}
    assert (
        json_digest(data)
        == "70c34029df6c2a201891cdc2f92cb4921dc7dbc5043b47b50813aad2aa4c4dc1"
    )


@pytest.mark.unit
def test_json_digest_key_order_is_irrelevant():
    """Key insertion order must not affect the digest."""
    a = {"z": 1, "a": 2}
    b = {"a": 2, "z": 1}
    assert json_digest(a) == json_digest(b)



"""
Tests for deep edges: extracting values from nested DataValue structures.

Tests run against both TopologicalExecutionAlgorithm and
ParallelExecutionAlgorithm via the `algorithm` fixture.
"""

from functools import cached_property
from typing import ClassVar, Literal

import pytest
from pydantic import Field

from workflow_engine import (
    Data,
    DataValue,
    Edge,
    Empty,
    ExecutionAlgorithm,
    IntegerValue,
    StringMapValue,
    StringValue,
    Workflow,
    WorkflowExecutionResultStatus,
)
from workflow_engine.contexts import InMemoryContext
from workflow_engine.core.io import InputNode, OutputNode
from workflow_engine.core.node import Node, NodeTypeInfo
from workflow_engine.execution import TopologicalExecutionAlgorithm
from workflow_engine.execution.parallel import ParallelExecutionAlgorithm


@pytest.fixture(params=["topological", "parallel"])
def algorithm(request) -> ExecutionAlgorithm:
    if request.param == "topological":
        return TopologicalExecutionAlgorithm()
    return ParallelExecutionAlgorithm()


class Level2Data(Data):
    """Deepest level."""

    alpha: IntegerValue = Field(title="Alpha", description="Deep integer.")
    beta: StringValue = Field(title="Beta", description="Deep string.")


class Level1Data(Data):
    """Middle level."""

    x: StringValue = Field(title="X", description="Mid-level string.")
    y: DataValue[Level2Data] = Field(title="Y", description="Nested level-2.")


class Level0Output(Data):
    """Top level with multiple nested DataValues."""

    a: IntegerValue = Field(title="A", description="Top-level integer.")
    b: DataValue[Level1Data] = Field(title="B", description="Nested level-1.")


class MultiLevelOutputNode(Node[Empty, Level0Output, Empty]):
    """Node that outputs a structure with multiple nesting levels."""

    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="MultiLevelOutput",
        display_name="Multi-Level Output",
        description="Outputs nested data for testing deep edges.",
        version="1.0.0",
        parameter_type=Empty,
    )

    type: Literal["MultiLevelOutput"] = "MultiLevelOutput"  # pyright: ignore[reportIncompatibleVariableOverride]
    params: Empty = Field(default_factory=Empty)

    @cached_property
    def input_type(self):
        return Empty

    @cached_property
    def output_type(self):
        return Level0Output

    async def run(self, context, input: Empty) -> Level0Output:
        return Level0Output(
            a=IntegerValue(100),
            b=DataValue[Level1Data](
                root=Level1Data(
                    x=StringValue("mid"),
                    y=DataValue[Level2Data](
                        root=Level2Data(
                            alpha=IntegerValue(7),
                            beta=StringValue("deep"),
                        ),
                    ),
                ),
            ),
        )


@pytest.mark.asyncio
async def test_deep_edges(algorithm: ExecutionAlgorithm):
    """
    Extract values at multiple nesting depths: leaf values, and whole nested
    DataValue objects.
    - 1 level:  ["a"] (leaf)
    - 2 levels: ["b", "x"] (leaf), ["b", "y"] (whole DataValue)
    - 3 levels: ["b", "y", "alpha"], ["b", "y", "beta"] (leaves)
    """
    node = MultiLevelOutputNode(id="multilevel")
    output_node = OutputNode.from_fields(
        top_num=IntegerValue,
        mid_str=StringValue,
        nested=DataValue[Level2Data],
        deep_num=IntegerValue,
        deep_str=StringValue,
    )

    workflow = Workflow(
        input_node=InputNode.from_fields(),
        output_node=output_node,
        inner_nodes=[node],
        edges=[
            Edge.from_nodes(
                source=node,
                source_key="a",
                target=output_node,
                target_key="top_num",
            ),
            Edge.from_nodes(
                source=node,
                source_key=["b", "x"],
                target=output_node,
                target_key="mid_str",
            ),
            Edge.from_nodes(
                source=node,
                source_key=["b", "y"],
                target=output_node,
                target_key="nested",
            ),
            Edge.from_nodes(
                source=node,
                source_key=["b", "y", "alpha"],
                target=output_node,
                target_key="deep_num",
            ),
            Edge.from_nodes(
                source=node,
                source_key=["b", "y", "beta"],
                target=output_node,
                target_key="deep_str",
            ),
        ],
    )

    context = InMemoryContext()
    result = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={},
    )

    assert result.status is WorkflowExecutionResultStatus.SUCCESS
    assert result.output["top_num"].root == 100
    assert result.output["mid_str"].root == "mid"
    assert isinstance(result.output["nested"], DataValue)
    assert result.output["nested"].root.alpha.root == 7
    assert result.output["nested"].root.beta.root == "deep"
    assert result.output["deep_num"].root == 7
    assert result.output["deep_str"].root == "deep"


# StringMapValue tests
class MapOutput(Data):
    """Output with a StringMapValue (dynamic keys)."""

    map: StringMapValue[IntegerValue] = Field(
        title="Map",
        description="A map of string keys to integer values.",
    )


class MapOutputNode(Node[Empty, MapOutput, Empty]):
    """Node that outputs a StringMapValue with known keys."""

    TYPE_INFO: ClassVar[NodeTypeInfo] = NodeTypeInfo.from_parameter_type(
        name="MapOutput",
        display_name="Map Output",
        description="Outputs a StringMapValue for testing dynamic key extraction.",
        version="1.0.0",
        parameter_type=Empty,
    )

    type: Literal["MapOutput"] = "MapOutput"  # pyright: ignore[reportIncompatibleVariableOverride]
    params: Empty = Field(default_factory=Empty)

    @cached_property
    def input_type(self):
        return Empty

    @cached_property
    def output_type(self):
        return MapOutput

    async def run(self, context, input: Empty) -> MapOutput:
        return MapOutput(
            map=StringMapValue[IntegerValue](
                root={"found": IntegerValue(42), "also_found": IntegerValue(99)}
            )
        )


@pytest.mark.asyncio
async def test_string_map_value_happy_path(algorithm: ExecutionAlgorithm):
    """Extract a value from StringMapValue when the key exists."""
    node = MapOutputNode(id="map_node")
    output_node = OutputNode.from_fields(num=IntegerValue)

    workflow = Workflow(
        input_node=InputNode.from_fields(),
        output_node=output_node,
        inner_nodes=[node],
        edges=[
            Edge.from_nodes(
                source=node,
                source_key=["map", "found"],
                target=output_node,
                target_key="num",
            ),
        ],
    )

    context = InMemoryContext()
    result = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={},
    )

    assert result.status is WorkflowExecutionResultStatus.SUCCESS
    assert result.output["num"].root == 42


@pytest.mark.asyncio
async def test_string_map_value_missing_key(algorithm: ExecutionAlgorithm):
    """Workflow fails when edge references a key that does not exist in StringMapValue."""
    node = MapOutputNode(id="map_node")
    output_node = OutputNode.from_fields(num=IntegerValue)

    workflow = Workflow(
        input_node=InputNode.from_fields(),
        output_node=output_node,
        inner_nodes=[node],
        edges=[
            Edge.from_nodes(
                source=node,
                source_key=["map", "missing_key"],
                target=output_node,
                target_key="num",
            ),
        ],
    )

    context = InMemoryContext()
    result = await algorithm.execute(
        context=context,
        workflow=workflow,
        input={},
    )

    assert result.status is WorkflowExecutionResultStatus.ERROR
    assert result.errors.any()



# workflow_engine/contexts/in_memory.py
from typing import TypeVar

from overrides import override

from ..core import Context, FileValue

F = TypeVar("F", bound=FileValue)


class InMemoryContext(Context):
    """
    Pretends to be a file system, but actually stores files in memory.
    """

    def __init__(self):
        self.data: dict[str, bytes] = {}

    @override
    async def read(
        self,
        file: FileValue,
    ) -> bytes:
        return self.data[file.path]

    @override
    async def write(
        self,
        file: F,
        content: bytes,
    ) -> F:
        self.data[file.path] = content
        return file


__all__ = [
    "InMemoryContext",
]



# tests/test_node_versioning.py
import warnings

import pytest
from pydantic import ValidationError

from workflow_engine import StringValue
from workflow_engine.nodes import ConstantStringNode
from workflow_engine.nodes.constant import ConstantString
from workflow_engine.utils.semver import LATEST_SEMANTIC_VERSION


class TestNodeVersioning:
    """Test node versioning functionality using ConstantStringNode."""

    @pytest.mark.unit
    def test_default_version_when_none_provided(self):
        """Test that when no version is provided, the node defaults to the current version."""
        # Create a node without specifying version
        node = ConstantStringNode(
            id="test_node",
            params=ConstantString(value=StringValue("test")),
        )

        # Should default to the current version from TYPE_INFO
        assert node.version == "0.4.0"
        assert node.version_tuple == (0, 4, 0)

    @pytest.mark.unit
    def test_serialization_includes_version(self):
        """Test that serializing the node includes the version."""
        # Create a node without specifying version
        node = ConstantStringNode(
            id="test_node",
            params=ConstantString(value=StringValue("test")),
        )

        # Serialize the node
        serialized = node.model_dump()

        # Should include the version
        assert "version" in serialized
        assert serialized["version"] == "0.4.0"

        # Should also work with JSON serialization
        json_str = node.model_dump_json()
        assert '"version":"0.4.0"' in json_str

    @pytest.mark.unit
    def test_older_version_triggers_warning(self):
        """Test that providing an older version triggers a warning."""
        # Create a node with an older version
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            node = ConstantStringNode(
                id="test_node",
                version="0.3.14",  # Older than current 0.4.0
                params=ConstantString(value=StringValue("test")),
            )

            # Should have triggered a warning
            assert len(w) == 1
            warning = w[0]
            assert issubclass(warning.category, UserWarning)

            warning_message = str(warning.message)
            assert (
                "Node version 0.3.14 is older than the latest version (0.4.0) supported by this workflow engine instance, and may need to be migrated."
                in warning_message
            )

            # Node should still be created successfully
            assert node.version == "0.3.14"
            assert node.version_tuple == (0, 3, 14)

    @pytest.mark.unit
    def test_newer_version_throws_error(self):
        """Test that providing a newer version throws an error."""
        with pytest.raises(ValidationError) as exc_info:
            ConstantStringNode(
                id="test_node",
                version="0.5.0",  # Newer than current 0.4.0
                params=ConstantString(value=StringValue("test")),
            )

        # Check that the error message is correct
        error_message = str(exc_info.value)
        assert (
            "Node version 0.5.0 is newer than the latest version (0.4.0) supported by this workflow engine instance."
            in error_message
        )

    @pytest.mark.unit
    def test_same_version_no_warning(self):
        """Test that providing the same version as current doesn't trigger warnings."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            node = ConstantStringNode(
                id="test_node",
                version="0.4.0",  # Same as current
                params=ConstantString(value=StringValue("test")),
            )

            # Should not have triggered any warnings
            assert len(w) == 0
            assert node.version == "0.4.0"

    @pytest.mark.unit
    def test_latest_version_constant(self):
        """Test that using LATEST_SEMANTIC_VERSION constant works correctly."""
        node = ConstantStringNode(
            id="test_node",
            version=LATEST_SEMANTIC_VERSION,
            params=ConstantString(value=StringValue("test")),
        )

        # Should default to the current version
        assert node.version == "0.4.0"
        assert node.version_tuple == (0, 4, 0)

    @pytest.mark.unit
    def test_version_validation_during_deserialization(self):
        """Test that version validation works during deserialization."""
        # Test with older version - should work but warn
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            node_data = {
                "id": "test_node",
                "type": "ConstantString",
                "version": "0.3.0",
                "params": {"value": "test"},
            }

            node = ConstantStringNode.model_validate(node_data)
            assert len(w) == 1  # Should have warned
            assert node.version == "0.3.0"

        # Test with newer version - should fail
        with pytest.raises(ValidationError):
            node_data = {
                "id": "test_node",
                "type": "ConstantString",
                "version": "0.5.0",
                "params": {"value": "test"},
            }
            ConstantStringNode.model_validate(node_data)
